do
local _ENV = _ENV
package.preload[ "modules.bot" ] = function( ... ) local arg = _G.arg;
local bot = {};

local script_handler;
local data_format;
local ljsocket;
local utils;

local obslua = obslua;
local require = require;
local assert = assert;
local string = string;
local print = print;
local error = error;
local tostring = tostring;

local host = "irc.chat.twitch.tv";
local port = 6667;
local bot_socket = nil;

local bot_nickname = "";
local bot_password = "";

local auth_success = false;
local auth_requested = false;

local joined_channel = "";

local password_length = 36;
local short_password_length = 30;

local oauth_string = "oauth:";

bot.is_timer_on = false;

function bot.tick()
	--print("Bot Tick");

	if not script_handler.is_script_enabled then
		return;
	end

	if bot_socket:is_connected() then
		if not auth_success and not auth_requested then
			bot.auth();
		end
		
		local response, err = bot.receive();

		if response ~= nil then
			for line in response:gmatch("[^\n]+") do
				if not auth_success then
					auth_requested = false;
					if line:match(":tmi.twitch.tv 001") then
						bot_nickname = bot.get_real_nickname(line);
						print("Authentication success: " .. bot_nickname);
						auth_success = true;
						
						bot.join_channel();
						goto continue;
					else 
						print("Authentication to " .. bot_nickname .. " failed! Socket closed! Trying to reconnect...");
						
						bot.reconnect();
						return;
					end
				end
				
				if line:match("PING") then
					bot.send("PONG");
					print("PING PONG");
					goto continue;
				end
				
				if line:match("JOIN") then
					print("Joined Channel: " .. joined_channel);
					goto continue;
				end
			
				local i = 0;
				local to_user = "";
				local command = "";
				
				for word in line:gmatch("[^%s]+") do
					if i == 0 then
						local j = 0;
						for token in word:gmatch("[^!]+") do
							if j == 0 then
								to_user = token:sub(2);
							end
							j = j + 1;
						end
						
					end
					
					if i == 1 then
						if word ~= "PRIVMSG" then
							return;
						end
					end
					
					if i == 3 then
						command = word:sub(2):lower();
					end
					
					if i == 4 then
						if not word:match("󠀀") then
							if word:match("^@") then
								to_user = word:sub(2);
							else
								to_user = word;
							end
						end
						
					end
					
					i = i + 1;
				end

				bot.process_commands(to_user, command);
				
				::continue::
			end
		end
	else
		bot_socket:poll_connect();
	end
end

function bot.process_commands(to_user, command)
	local formatted_stats = data_format.stats;

	if command:match("^!encoder") then

		bot.send_message(string.format("@%s -> Encoder: %s",
			to_user,
			formatted_stats.encoder
		));

	elseif command:match("^!output_mode") or command:match("^!outputmode") then

		bot.send_message(string.format("@%s -> Output Mode: %s",
			to_user,
			formatted_stats.output_mode
		));
		
	elseif command:match("^!canvas_resolution") or command:match("^!canvasresolution") then

		bot.send_message(string.format("@%s -> Canvas Resolution: %s",
			to_user,
			formatted_stats.canvas_resolution
		));
		
	elseif command:match("^!output_resolution") or command:match("^!outputresolution") then

		bot.send_message(string.format("@%s -> Output Resolution: %s",
			to_user,
			formatted_stats.output_resolution
		));
		
	elseif command:match("^!missed_frames") or command:match("^!missedframes") or command:match("^!missed") then

		bot.send_message(string.format("@%s -> Missed Frames: %s/%s (%s%%)",
			to_user,
			formatted_stats.missed_frames,
			formatted_stats.total_missed_frames,
			formatted_stats.missed_percents
		));
		
	elseif command:match("^!skipped_frames") or command:match("^!skippedframes") or command:match("^!skipped") then

		bot.send_message(string.format("@%s -> Skipped Frames: %s/%s (%s%%)",
			to_user,
			formatted_stats.skipped_frames,
			formatted_stats.total_skipped_frames,
			formatted_stats.skipped_percents
		));
		
	elseif command:match("^!dropped_frames") or command:match("^!droppedframes") or command:match("^!dropped") then

		bot.send_message(string.format("@%s -> Dropped Frames: %s/%s (%s%%)",
			to_user,
			formatted_stats.dropped_frames,
			formatted_stats.total_dropped_frames,
			formatted_stats.dropped_percents
		));
		
	elseif command:match("^!congestion") then

		bot.send_message(string.format("@%s -> Congestion: %s%% (Average: %s%%)",
			to_user,
			formatted_stats.congestion,
			formatted_stats.average_congestion
		));
		
	elseif command:match("^!frame_time") or command:match("^!render_time") or command:match("^!frametime") or command:match("^!rendertime") then

		bot.send_message(string.format("@%s -> Average Frame Time: %s ms",
			to_user,
			formatted_stats.average_frame_time
		));
		
	elseif command:match("^!fps") or command:match("^!framerate") then

		bot.send_message(string.format("@%s -> FPS: %s/%s (Average: %s)",
			to_user,
			formatted_stats.fps,
			formatted_stats.target_fps,
			formatted_stats.average_fps
		));
		
	elseif command:match("^!memory_usage") or command:match("^!memoryusage") or command:match("^!memory") then

		bot.send_message(string.format("@%s -> Memory Usage: %s MB",
			to_user,
			formatted_stats.memory_usage
		));
		
	elseif command:match("^!cpu_cores") or command:match("^!cpucores") or command:match("^!cores") then

		bot.send_message(string.format("@%s -> CPU Cores: %s",
			to_user,
			formatted_stats.cpu_cores
		));

	elseif command:match("^!cpu_usage") or command:match("^!cpuusage") then

		bot.send_message(string.format("@%s -> CPU Usage: %s%%",
			to_user,
			formatted_stats.cpu_usage
		));

	elseif command:match("^!audio_bitrate") or command:match("^!audiobitrate") then

		bot.send_message(string.format("@%s -> Audio Bitrate: %s kb/s",
			to_user,
			formatted_stats.audio_bitrate
		));

	elseif command:match("^!bitrate") then

		bot.send_message(string.format("@%s -> Bitrate: %s kb/s",
			to_user,
			formatted_stats.bitrate
		));

	elseif command:match("^!recording_bitrate") or command:match("^!recordingbitrate")then

		bot.send_message(string.format("@%s -> Recording Bitrate: %s kb/s",
			to_user,
			formatted_stats.recording_bitrate
		));

	elseif command:match("^!streaming_duration") or command:match("^!streamingduration") then

		bot.send_message(string.format("@%s -> Streaming Duration: %s",
			to_user, 
			formatted_stats.streaming_duration
		));
	
	elseif command:match("^!recording_duration") or command:match("^!recordingduration") then

		bot.send_message(string.format("@%s -> Recording Duration: %s",
			to_user,
			formatted_stats.recording_duration
		));

	elseif command:match("^!streaming_status") or command:match("^!streamingstatus") then

		bot.send_message(string.format("@%s -> Streaming Status: %s",
			to_user, 
			formatted_stats.streaming_status
		));

	elseif command:match("^!recording_status") or command:match("^!recordingstatus") then

		bot.send_message(string.format("@%s -> Recording Status: %s",
			to_user,
			formatted_stats.recording_status
		));
		
	elseif command:match("^!obs_static_stats") or command:match("^!obsstaticstats") then

		bot.send_message(string.format("@%s -> Encoder: %s, Output Mode: %s, Canvas Resolution: %s, Output Resolution: %s, CPU Cores: %s, Audio Bitrate: %s kb/s",
			to_user,
			formatted_stats.encoder,
			formatted_stats.output_mode,
			formatted_stats.canvas_resolution,
			formatted_stats.output_resolution,
			formatted_stats.cpu_cores,
			formatted_stats.audio_bitrate
		));

	elseif command:match("^!obs_stats") or command:match("^!obsstats") or command:match("^!obs_dynamic_stats") or command:match("^!obsdynamicstats") then

		bot.send_message(string.format("@%s -> Missed: %s/%s (%s%%), Skipped: %s/%s (%s%%), Dropped: %s/%s (%s%%), Cong.: %s%% (average: %s%%), Frame Time: %s ms, FPS: %s/%s (average: %s), RAM: %s MB, CPU: %s%%, Bitrate: %s kb/s",
			to_user,
			formatted_stats.missed_frames,
			formatted_stats.total_missed_frames,
			formatted_stats.missed_percents,
			formatted_stats.skipped_frames,
			formatted_stats.total_skipped_frames,
			formatted_stats.skipped_percents,
			formatted_stats.dropped_frames,
			formatted_stats.total_dropped_frames,
			formatted_stats.dropped_percents,
			formatted_stats.congestion,
			formatted_stats.average_congestion,
			formatted_stats.average_frame_time,
			formatted_stats.fps,
			formatted_stats.target_fps,
			formatted_stats.average_fps,
			formatted_stats.memory_usage,
			formatted_stats.cpu_usage,
			formatted_stats.bitrate
		));
	end
end

function bot.auth()
	print("Authentication attempt: " .. bot_nickname);
	auth_requested = true;
	assert(bot_socket:send(
		string.format("PASS %s\r\nNICK %s\r\n", bot_password, bot_nickname)
	));
end

function bot.send(message)
	if message ~= "PONG" then
		print(string.format("Sending Message: %s", message));
	end

	assert(bot_socket:send(
		string.format("%s\r\n", message)
	));
end

function bot.send_message(message)
	print(string.format("Sending Message to %s: %s", joined_channel, message));
	assert(bot_socket:send(
		string.format("PRIVMSG #%s :%s\r\n", joined_channel, message)
	));
end

function bot.receive()
	local response, err = bot_socket:receive();

	if response ~= nil then
		return response;
	elseif err ~= nil then
		if err == "timeout" then
			return nil;
		--"An established connection was aborted by the software in your host machine."
		elseif err:match("An established connection was aborted") then
			print(tostring(err));
			bot.reconnect();
			return nil;
		else
			print(tostring(err));
			return nil;
		end
	else
		print("Unknown Error");
		return nil;
	end
end

function bot.join_channel(channel)
	if channel == nil or channel == "" then
		channel = script_handler.channel_nickname;
	end

	if channel == nil or channel == "" then
		channel = bot_nickname;
	end

	joined_channel = channel;
	bot.send("JOIN #" .. channel);

end

function bot.get_real_nickname(line)
	local i = 0;
	for word in line:gmatch("[^%s]+") do
		if i == 2 then
			return word;
		end
		i = i + 1;
	end
end

function bot.reconnect()
	print("Reconnecting...");
	
	bot.close_socket();
	bot.init_socket();
end
	
function bot.init_socket()
	if not script_handler.is_bot_enabled then
		return;
	end

	local nickname = bot.validate_nickname(script_handler.bot_nickname);
	local password = bot.validate_password(script_handler.bot_password);

	if nickname ~= nil then
		bot_nickname = nickname;
	else
		return;
	end

	if password ~= nil then
		bot_password = password;
	else
		return;
	end

	bot_socket = assert(ljsocket.create("inet", "stream", "tcp"));
	assert(bot_socket:set_blocking(false));
	assert(bot_socket:connect(host, port));

	if not bot.is_timer_on then
		bot.is_timer_on = true;
		obslua.timer_add(bot.tick, script_handler.bot_delay);

		print("Bot Timer added.");
	end
end

function bot.close_socket()
	if bot.is_timer_on then
		bot.is_timer_on = false;
		obslua.timer_remove(bot.tick);

		print("Bot Timer removed.");
	end

	if bot_socket ~= nil then
		bot_socket:close();
		bot_socket = nil;
	end
	
	bot.reset_data();
end

function bot.close_socket_on_unload()
	if bot_socket ~= nil then
		bot_socket:close();
	end
end

function bot.validate_nickname(nickname)
	if nickname == nil then
		print("No Bot Nickname provided.");
		return nil;
	end

	nickname = utils.trim(nickname):lower();

	if nickname == "" then
		print("No Bot Nickname provided.");
		return nil;
	end

	return nickname;
end

function bot.validate_password(password)
	if password == nil then
		print("No Bot OAuth Password provided.");
		return nil;
	end

	password = utils.trim(password):lower();

	if password == "" then
		print("No Bot OAuth Password provided.");
		return nil;
	end

	if utils.starts_with(password, oauth_string) then
		if #password ~= password_length then
			print("Incorrect OAuth Password provided.");
			return nil;
		end
	else
		if #password ~= short_password_length then
			print("Incorrect OAuth Password provided.");
			return nil;
		end

		password = oauth_string .. password;
	end

	return password;
end

function bot.reset_data()
	auth_success = false;
	auth_requested = false;
end

function bot.init_module()
	script_handler = require("modules.script_handler");
	data_format = require("modules.data_format");
	ljsocket = require("modules.ljsocket");
	utils = require("modules.utils");
end

return bot;
end
end

do
local _ENV = _ENV
package.preload[ "modules.data" ] = function( ... ) local arg = _G.arg;
local data = {};

local ffi;
local text_source_handler;
local utils;

local obslua = obslua;
local print = print;
local tostring = tostring;
local math = math;
local require = require;
local string = string;
local os = os;

local obsffi = nil;
local bytes_to_megabytes = 1024 * 1024;
local ns_to_ms = 1000000;
local ms_to_s = 1000;
local s_to_m = 60;
local s_to_h = 3600;
local cpu_info  = nil;
local bitrate_update_delay = 2000;

local default_encoder = "x264"
local default_canvas_width = 1920;
local default_canvas_resolution = "1920x1080";
local default_output_resolution = "1280x720";
local default_target_fps = 30;

local defaults = {
	encoder = "x264?",
	output_mode = "Simple",
	canvas_width = 1920,
	canvas_height = 1080,
	canvas_resolution = "1920x1080?",
	output_width = 1280,
	output_height = 720,
	output_resolution = "1280x720",
	target_fps = 30,
	audio_bitrate = 160
}

local cumulative_fps = 0;
local cumulative_congestion = 0;
local last_bitrate_update_time = 0;
local last_bytes_sent = 0;
local last_bytes_recorded = 0;

data.ticks = 0;

data.streaming_statuses = {
	live = "Live",
	reconnecting = "Reconnecting",
	offline = "Offline"
};

data.recording_statuses = {
	on = "On",
	paused = "Paused",
	off = "Off"
};

data.output_modes = {
	simple = "Simple",
	advanced = "Advanced"
};

data.stats = {
	date = "01.01.1970",
	time = "00:00:00",
	date_time = "01.01.1970 00:00:00",
	encoder = defaults.encoder,
	output_mode = defaults.output_mode,
	canvas_width = defaults.canvas_width,
	canvas_height = defaults.canvas_height,
	canvas_resolution = defaults.canvas_resolution,
	output_width = defaults.output_width,
	output_height = defaults.output_height,
	output_resolution = defaults.output_resolution,
	missed_frames = 0,
	total_missed_frames = 0,
	missed_percents = 0,
	skipped_frames = 0,
	total_skipped_frames = 0,
	skipped_percents = 0,
	dropped_frames = 0,
	total_dropped_frames = 0,
	dropped_percents = 0,
	congestion = 0,
	average_congestion = 0,
	average_frame_time = 0,
	fps = 0,
	target_fps = defaults.target_fps,
	average_fps = 0,
	memory_usage = 0,
	cpu_physical_cores = 0,
	cpu_logical_cores = 0,
	cpu_cores = "0C/0T",
	cpu_usage = 0,
	audio_bitrate = defaults.audio_bitrate,
	recording_bitrate = 0,
	bitrate = 0,
	streaming_total_seconds = 0,
	streaming_total_minutes = 0,
	streaming_hours = 0,
	streaming_minutes = 0,
	streaming_seconds = 0;
	streaming_duration = "00:00:00",
	recording_total_seconds = 0,
	recording_total_minutes = 0,
	recording_hours = 0,
	recording_minutes = 0,
	recording_seconds = 0;
	recording_duration = "00:00:00",
	streaming_status = "Offline",
	recording_status = "Off"
}

function data.update()
	--print("Data Update Tick.");

	data.ticks = data.ticks + 1;

	-- streaming_output will be nil when not actually streaming
	local streaming_output = obslua.obs_frontend_get_streaming_output();
	local recording_output = obslua.obs_frontend_get_recording_output();

	local bitrate_time_passed = data.update_bitrate_time_passed();

	data.update_time();
	data.update_cpu_usage();
	data.update_memory_usage();
	data.update_fps();
	data.update_average_frame_time();
	data.update_missed_frames();
	data.update_skipped_frames();
	data.update_dropped_frames(streaming_output);
	data.update_congestion(streaming_output);
	data.update_reconnection_status(streaming_output);
	data.update_streaming_bitrate(streaming_output, bitrate_time_passed);
	data.update_streaming_duration();
	data.update_recording_bitrate(recording_output, bitrate_time_passed);
	data.update_recording_duration(recording_output);

	if streaming_output ~= nil then
		obslua.obs_output_release(streaming_output);
	end

	if recording_output ~= nil then
		obslua.obs_output_release(recording_output);
	end
end

function data.update_time()
	data.stats.date = os.date("%d.%m.%Y");
	data.stats.time = os.date("%X");
	data.stats.date_time = os.date("%d.%m.%Y %X");
end

function data.update_bitrate_time_passed()
	local current_time = obslua.os_gettime_ns();
	local time_passed = (current_time - last_bitrate_update_time) / ns_to_ms;
	
	if time_passed >= bitrate_update_delay then
		last_bitrate_update_time = current_time;
	end

	return time_passed;
end

function data.start_cpu_usage_info()
	if obsffi == nil then
		return;
	end

	data.destroy_cpu_usage_info();

	data.cpu_info = obsffi.os_cpu_usage_info_start();
end

function data.destroy_cpu_usage_info()
	if data.cpu_info == nil or obsffi == nil then
		return;
	end

	obsffi.os_cpu_usage_info_destroy(data.cpu_info);
	data.cpu_info = nil;
end

function data.update_cpu_usage()
	if data.cpu_info == nil or obsffi == nil then
		return;
	end
	
	local cpu_usage = obsffi.os_cpu_usage_info_query(data.cpu_info);

	if cpu_usage ~= nil then
		data.stats.cpu_usage = cpu_usage / 100;
	end
	
end

function data.update_memory_usage()
	local memory_usage = obslua.os_get_proc_resident_size() / bytes_to_megabytes;

	if memory_usage ~= nil then
		data.stats.memory_usage = memory_usage;
	end
end

function data.update_fps()
	local fps = obslua.obs_get_active_fps();

	if fps ~= nil then
		data.stats.fps = fps;
		cumulative_fps = cumulative_fps + fps;
		data.stats.average_fps = cumulative_fps / data.ticks;
	end
end

function data.update_average_frame_time()
	local average_frame_time = obslua.obs_get_average_frame_time_ns() / ns_to_ms;

	if average_frame_time ~= nil then
		data.stats.average_frame_time = average_frame_time;
	end
end

function data.update_missed_frames()
	local total_missed_frames = obslua.obs_get_total_frames();	-- total rendered frames
	local missed_frames = obslua.obs_get_lagged_frames();		-- lagged frames

	if total_missed_frames ~= nil then
		data.stats.total_missed_frames = total_missed_frames;
	end

	if missed_frames ~= nil then
		data.stats.missed_frames = missed_frames;
	end

	if data.stats.total_missed_frames == 0 then
		data.stats.missed_percents = 0;
	else
		data.stats.missed_percents = data.stats.missed_frames / data.stats.total_missed_frames;
	end
end

function data.update_skipped_frames()
	if obsffi == nil then
		return;
	end

	local video = obsffi.obs_get_video();

	if video == nil then
		return;
	end

	local total_skipped_frames = obsffi.video_output_get_total_frames(video);	-- total encoded frames
	local skipped_frames = obsffi.video_output_get_skipped_frames(video);		-- skipped frames

	if total_skipped_frames ~= nil then
		data.stats.total_skipped_frames = total_skipped_frames;
	end

	if skipped_frames ~= nil then
		data.stats.skipped_frames = skipped_frames;
	end

	if data.stats.total_skipped_frames == 0 then
		data.stats.skipped_percents = 0;
	else
		data.stats.skipped_percents = data.stats.skipped_frames / data.stats.total_skipped_frames;
	end
end

function data.update_dropped_frames(streaming_output)
	if streaming_output == nil then
		return;
	end
	
	local dropped_frames = obslua.obs_output_get_frames_dropped(streaming_output); -- dropped frames
	local total_frames = obslua.obs_output_get_total_frames(streaming_output);		-- total dropped frames
	
	if total_frames ~= nil then
		data.stats.total_dropped_frames = total_frames;
	end

	if dropped_frames ~= nil then
		data.stats.dropped_frames = dropped_frames;
	end

	if data.stats.total_dropped_frames == 0 then
		data.stats.dropped_percents = 0;
	else
		data.stats.dropped_percents = data.stats.dropped_frames / data.stats.total_dropped_frames;
	end
end

function data.update_congestion(streaming_output)
	if streaming_output == nil then
		return;
	end
	
	local congestion = obslua.obs_output_get_congestion(streaming_output);

	-- Check that congestion is not NaN
	if(congestion ~= nil and not utils.is_NaN(congestion)) then
		cumulative_congestion = cumulative_congestion + congestion;

		data.stats.congestion = congestion;
		data.stats.average_congestion =  cumulative_congestion / data.ticks;
	end
end

function data.update_reconnection_status(streaming_output)
	if streaming_output == nil then
		return;
	end

	if data.stats.streaming_status == data.streaming_statuses.live
	or data.stats.streaming_status == data.streaming_statuses.reconnecting then

		local is_reconnecting = obslua.obs_output_reconnecting(streaming_output);
		if is_reconnecting then
			data.stats.streaming_status = data.streaming_statuses.reconnecting;
		else
			data.stats.streaming_status = data.streaming_statuses.live;
		end
	end
end

function data.update_streaming_status(is_live)
	if is_live then
		data.stats.streaming_status = data.streaming_statuses.live;
	else
		data.stats.streaming_status = data.streaming_statuses.offline;
		data.stats.bitrate = 0;
	end
end

function data.update_recording_status(is_recording, is_paused)
	if is_recording then
		if is_paused then
			data.stats.recording_status = data.recording_statuses.paused;
		else
			data.stats.recording_status = data.recording_statuses.on;
		end
		
	else
		data.stats.recording_status = data.recording_statuses.off;
		data.stats.recording_bitrate = 0;
	end
end

function data.update_streaming_bitrate(streaming_output, time_passed)
	if streaming_output == nil then
		return;
	end

	if time_passed < bitrate_update_delay then
		return;
	end

	local bytes_sent = obslua.obs_output_get_total_bytes(streaming_output);

	if bytes_sent == nil then
		return;
	end
	
	-- the fck is this?
	if bytes_sent < last_bytes_sent then
		bytes_sent = 0;
	end
	
	local bits_between = (bytes_sent - last_bytes_sent) * 8;
	local bitrate = bits_between / time_passed;

	last_bytes_sent = bytes_sent;

	if bitrate ~= nil then
		data.stats.bitrate = bitrate;
	end
end

function data.update_recording_bitrate(recording_output, time_passed)
	if recording_output == nil then
		return;
	end

	if time_passed < bitrate_update_delay then
		return;
	end
	
	local bytes_recorded = obslua.obs_output_get_total_bytes(recording_output);

	if bytes_recorded == nil then
		return;
	end
	
	-- what the fck is this?
	if bytes_recorded < last_bytes_recorded then
		bytes_recorded = 0;
	end

	local recording_bits_between = (bytes_recorded - last_bytes_recorded) * 8;
	local recording_bitrate = recording_bits_between / time_passed / ms_to_s;

	last_bytes_recorded = bytes_recorded;

	if recording_bitrate ~= nil then
		data.stats.recording_bitrate = recording_bitrate;
	end
end

function data.update_streaming_duration()
	-- Needs better approach?
	-- Duration is incorrect if fps is not stable

	if data.stats.streaming_status == data.streaming_statuses.offline then
		data.stats.streaming_total_minutes = 0;
		data.stats.streaming_total_seconds = 0;
		data.stats.streaming_hours = 0;
		data.stats.streaming_minutes = 0;
		data.stats.streaming_seconds = 0;
		data.stats.streaming_duration = "00:00:00";

		return;
	end

	local fps = data.stats.fps;

	local streaming_total_seconds = data.stats.total_dropped_frames;
	local streaming_total_minutes = 0;
	local streaming_hours = 0;
	local streaming_minutes = 0;
	local streaming_seconds = 0;

	if fps ~= 0 then
		streaming_total_seconds = streaming_total_seconds / fps;
	end

	streaming_hours = math.floor(streaming_total_seconds / 3600);
	streaming_minutes = math.floor((streaming_total_seconds % 3600) / 60);
	streaming_seconds = math.floor(0.5 + streaming_total_seconds % 60);

	data.stats.streaming_total_minutes = math.floor(streaming_total_seconds / 60);
	data.stats.streaming_total_seconds = streaming_total_seconds;

	data.stats.streaming_hours = streaming_hours;
	data.stats.streaming_minutes = streaming_minutes;
	data.stats.streaming_seconds = streaming_seconds;
	data.stats.streaming_duration = string.format("%.2d:%.2d:%.2d", streaming_hours, streaming_minutes, streaming_seconds);
end

function data.update_recording_duration(recording_output)
	-- Needs better approach?
	-- Duration is incorrect if fps is not stable

	if recording_output == nil then
		return;
	end

	if data.stats.recording_status == data.recording_statuses.off then
		data.stats.recording_total_minutes = 0;
		data.stats.recording_total_seconds = 0;
		data.stats.recording_hours = 0;
		data.stats.recording_minutes = 0;
		data.stats.recording_seconds = 0;
		data.stats.recording_duration = "00:00:00";

		return;
	end

	local recording_total_frames = obslua.obs_output_get_total_frames(recording_output);

	if recording_total_frames == nil then
		return;
	end

	local fps = data.stats.fps;

	local recording_total_seconds = recording_total_frames;
	local recording_total_minutes = 0;
	local recording_hours = 0;
	local recording_minutes = 0;
	local recording_seconds = 0;

	if fps ~= 0 then
		recording_total_seconds = recording_total_seconds / fps;
	end

	recording_hours = math.floor(recording_total_seconds / 3600);
	recording_minutes = math.floor((recording_total_seconds % 3600) / 60);
	recording_seconds = math.floor(0.5 + recording_total_seconds % 60);

	data.stats.recording_total_minutes = math.floor(recording_total_seconds / 60);
	data.stats.recording_total_seconds = recording_total_seconds;

	data.stats.recording_hours = recording_hours;
	data.stats.recording_minutes = recording_minutes;
	data.stats.recording_seconds = recording_seconds;
	data.stats.recording_duration = string.format("%.2d:%.2d:%.2d", recording_hours, recording_minutes, recording_seconds);
end

function data.update_output_mode(output_mode)
	if output_mode == nil then
		data.stats.output_mode = defaults.output_mode;
	else
		data.stats.output_mode = output_mode;
	end
end

function data.update_encoder(encoder)
	if encoder ~= nil then
		data.stats.encoder = encoder;
	end
end

function data.update_canvas_resolution(width, height)
	if width ~= nil then
		data.stats.canvas_width = width;
	end

	if height ~= nil then
		data.stats.canvas_height = height;
	end

	if width ~= nil or height ~= nil then
		data.stats.canvas_resolution = string.format("%dx%d", data.stats.canvas_width, data.stats.canvas_height);
	end
end

function data.update_output_resolution(width, height)
	if width == nil then
		data.stats.output_width = defaults.output_width;
	else
		data.stats.output_width = width;
	end

	if height == nil then
		data.stats.output_height = defaults.output_height;
	else
		data.stats.output_height = height;
	end

	data.stats.output_resolution = string.format("%dx%d", data.stats.output_width, data.stats.output_height);
end

function data.update_target_fps(target_fps)
	if target_fps == nil then
		data.stats.target_fps = defaults.target_fps;
	else
		data.stats.target_fps = target_fps;
	end
end

function data.update_audio_bitrate(audio_bitrate)
	if audio_bitrate == nil then
		data.stats.audio_bitrate = defaults.audio_bitrate;
	else
		data.stats.audio_bitrate = audio_bitrate;
	end
end

function data.update_cores_on_script_settings_changed()
	local physical_cores = obslua.os_get_physical_cores();
	local logical_cores = obslua.os_get_logical_cores();

	if physical_cores ~= nil then
		data.stats.cpu_physical_cores = physical_cores;
	end

	if logical_cores ~= nil then
		data.stats.cpu_logical_cores = logical_cores;
	end

	data.stats.cpu_cores = string.format("%dC/%dT", data.stats.cpu_physical_cores, data.stats.cpu_logical_cores);
end

function data.update_streaming_status_on_script_settings_changed()
	local is_streaming_active = obslua.obs_frontend_streaming_active();

	if is_streaming_active == nil then
		return;
	end

	if is_streaming_active then
		data.stats.streaming_status = data.streaming_statuses.live;
	else
		data.stats.streaming_status = data.streaming_statuses.offline;
		data.stats.bitrate = 0;
	end
end

function data.update_recording_status_on_script_settings_changed()
	local is_recording_active = obslua.obs_frontend_recording_active();

	if is_recording_active == nil then
		return;
	end

	if is_recording_active then
		local is_recording_paused = obslua.obs_frontend_recording_paused();

		if is_recording_paused == nil then
			return;
		end

		if is_recording_paused then
			data.stats.recording_status = data.recording_statuses.paused;
		else 
			data.stats.recording_status = data.recording_statuses.on;
		end
	else
		data.stats.recording_status = data.recording_statuses.off;
		data.stats.recording_bitrate = 0;
	end
end

function data.init_module()
	ffi = require("ffi");
	text_source_handler = require("modules.text_source_handler");
	utils = require("modules.utils");
	
	ffi.cdef[[
		struct video_output;
		typedef struct video_output video_t;

		struct os_cpu_usage_info;
		typedef struct os_cpu_usage_info os_cpu_usage_info_t;

		uint32_t video_output_get_skipped_frames(const video_t *video);
		uint32_t video_output_get_total_frames(const video_t *video);
		double video_output_get_frame_rate(const video_t *video);
		
		os_cpu_usage_info_t *os_cpu_usage_info_start(void);
		double os_cpu_usage_info_query(os_cpu_usage_info_t *info);
		void os_cpu_usage_info_destroy(os_cpu_usage_info_t *info);

		video_t *obs_get_video(void);
	]]

	if ffi.os == "OSX" then
		obsffi = ffi.load("obs-opengl.dylib"); -- OS X
	else
		obsffi = ffi.load("obs"); -- Windows
		-- Linux?
	end
end

return data;
end
end

do
local _ENV = _ENV
package.preload[ "modules.data_format" ] = function( ... ) local arg = _G.arg;
local data_format = {};

local data;

local require = require;
local tostring = tostring;
local string = string;

data_format.stats = {};

function data_format.update()
	local stats = data.stats;

	data_format.stats.date_time = tostring(stats.date_time);
	data_format.stats.date = tostring(stats.date);
	data_format.stats.time = tostring(stats.time);

	data_format.stats.encoder = tostring(stats.encoder);
	data_format.stats.output_mode = tostring(stats.output_mode);

	data_format.stats.canvas_width = string.format("%d", stats.canvas_width);
	data_format.stats.canvas_height = string.format("%d", stats.canvas_height);
	data_format.stats.canvas_resolution = tostring(stats.canvas_resolution);

	data_format.stats.output_width = string.format("%d", stats.output_width);
	data_format.stats.output_height = string.format("%d", stats.output_height);
	data_format.stats.output_resolution = tostring(stats.output_resolution);

	data_format.stats.missed_frames = string.format("%d", stats.missed_frames);
	data_format.stats.total_missed_frames = string.format("%d", stats.total_missed_frames);
	data_format.stats.missed_percents = string.format("%.2f", 100 * stats.missed_percents);

	data_format.stats.skipped_frames = string.format("%d", stats.skipped_frames);
	data_format.stats.total_skipped_frames = string.format("%d", stats.total_skipped_frames);
	data_format.stats.skipped_percents = string.format("%.2f", 100 * stats.skipped_percents);

	data_format.stats.dropped_frames = string.format("%d", stats.dropped_frames);
	data_format.stats.total_dropped_frames = string.format("%d", stats.total_dropped_frames);
	data_format.stats.dropped_percents = string.format("%.2f", 100 * stats.dropped_percents);

	data_format.stats.congestion = string.format("%.2f", 100 * stats.congestion);
	data_format.stats.average_congestion = string.format("%.2f", 100 * stats.average_congestion);

	data_format.stats.average_frame_time = string.format("%.1f", stats.average_frame_time);
	data_format.stats.fps = string.format("%.2f", stats.fps);
	data_format.stats.target_fps = string.format("%d", stats.target_fps);
	data_format.stats.average_fps = string.format("%.2f", stats.average_fps);

	data_format.stats.memory_usage = string.format("%.1f", stats.memory_usage);
	data_format.stats.cpu_physical_cores = string.format("%d", stats.cpu_physical_cores);
	data_format.stats.cpu_logical_cores = string.format("%d", stats.cpu_logical_cores);
	data_format.stats.cpu_cores = tostring(stats.cpu_cores);
	data_format.stats.cpu_usage = string.format("%.2f", 100 * stats.cpu_usage);

	data_format.stats.audio_bitrate = string.format("%d", stats.audio_bitrate);
	data_format.stats.recording_bitrate = string.format("%d", stats.recording_bitrate);
	data_format.stats.bitrate = string.format("%d", stats.bitrate);

	data_format.stats.streaming_total_seconds = string.format("%d", stats.streaming_total_seconds);
	data_format.stats.streaming_total_minutes = string.format("%d", stats.streaming_total_minutes);
	data_format.stats.streaming_hours = string.format("%d", stats.streaming_hours);
	data_format.stats.streaming_minutes = string.format("%d", stats.streaming_minutes);
	data_format.stats.streaming_seconds = string.format("%d", stats.streaming_seconds);
	data_format.stats.streaming_duration = tostring(stats.streaming_duration);

	data_format.stats.recording_total_seconds = string.format("%d", stats.recording_total_seconds);
	data_format.stats.recording_total_minutes = string.format("%d", stats.recording_total_minutes);
	data_format.stats.recording_hours = string.format("%d", stats.recording_hours);
	data_format.stats.recording_minutes = string.format("%d", stats.recording_minutes);
	data_format.stats.recording_seconds = string.format("%d", stats.recording_seconds);
	data_format.stats.recording_duration = tostring(stats.recording_duration);

	data_format.stats.streaming_status = tostring(stats.streaming_status);
	data_format.stats.recording_status = tostring(stats.recording_status);
end

function data_format.init_module()
	data = require("modules.data");
end

return data_format;
end
end

do
local _ENV = _ENV
package.preload[ "modules.json" ] = function( ... ) local arg = _G.arg;
-- Module options:
local always_use_lpeg = false
local register_global_module_table = false
local global_module_name = 'json'

--[==[

David Kolf's JSON module for Lua 5.1 - 5.4

Version 2.6


For the documentation see the corresponding readme.txt or visit
<http://dkolf.de/src/dkjson-lua.fsl/>.

You can contact the author by sending an e-mail to 'david' at the
domain 'dkolf.de'.


Copyright (C) 2010-2021 David Heiko Kolf

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

--]==]

-- global dependencies:
local pairs, type, tostring, tonumber, getmetatable, setmetatable, rawset =
      pairs, type, tostring, tonumber, getmetatable, setmetatable, rawset
local error, require, pcall, select = error, require, pcall, select
local floor, huge = math.floor, math.huge
local strrep, gsub, strsub, strbyte, strchar, strfind, strlen, strformat =
      string.rep, string.gsub, string.sub, string.byte, string.char,
      string.find, string.len, string.format
local strmatch = string.match
local concat = table.concat

local json = { version = "dkjson 2.6" }

local jsonlpeg = {}

if register_global_module_table then
  if always_use_lpeg then
    _G[global_module_name] = jsonlpeg
  else
    _G[global_module_name] = json
  end
end

local _ENV = nil -- blocking globals in Lua 5.2 and later

pcall (function()
  -- Enable access to blocked metatables.
  -- Don't worry, this module doesn't change anything in them.
  local debmeta = require "debug".getmetatable
  if debmeta then getmetatable = debmeta end
end)

json.null = setmetatable ({}, {
  __tojson = function () return "null" end
})

local function isarray (tbl)
  local max, n, arraylen = 0, 0, 0
  for k,v in pairs (tbl) do
    if k == 'n' and type(v) == 'number' then
      arraylen = v
      if v > max then
        max = v
      end
    else
      if type(k) ~= 'number' or k < 1 or floor(k) ~= k then
        return false
      end
      if k > max then
        max = k
      end
      n = n + 1
    end
  end
  if max > 10 and max > arraylen and max > n * 2 then
    return false -- don't create an array with too many holes
  end
  return true, max
end

local escapecodes = {
  ["\""] = "\\\"", ["\\"] = "\\\\", ["\b"] = "\\b", ["\f"] = "\\f",
  ["\n"] = "\\n",  ["\r"] = "\\r",  ["\t"] = "\\t"
}

local function escapeutf8 (uchar)
  local value = escapecodes[uchar]
  if value then
    return value
  end
  local a, b, c, d = strbyte (uchar, 1, 4)
  a, b, c, d = a or 0, b or 0, c or 0, d or 0
  if a <= 0x7f then
    value = a
  elseif 0xc0 <= a and a <= 0xdf and b >= 0x80 then
    value = (a - 0xc0) * 0x40 + b - 0x80
  elseif 0xe0 <= a and a <= 0xef and b >= 0x80 and c >= 0x80 then
    value = ((a - 0xe0) * 0x40 + b - 0x80) * 0x40 + c - 0x80
  elseif 0xf0 <= a and a <= 0xf7 and b >= 0x80 and c >= 0x80 and d >= 0x80 then
    value = (((a - 0xf0) * 0x40 + b - 0x80) * 0x40 + c - 0x80) * 0x40 + d - 0x80
  else
    return ""
  end
  if value <= 0xffff then
    return strformat ("\\u%.4x", value)
  elseif value <= 0x10ffff then
    -- encode as UTF-16 surrogate pair
    value = value - 0x10000
    local highsur, lowsur = 0xD800 + floor (value/0x400), 0xDC00 + (value % 0x400)
    return strformat ("\\u%.4x\\u%.4x", highsur, lowsur)
  else
    return ""
  end
end

local function fsub (str, pattern, repl)
  -- gsub always builds a new string in a buffer, even when no match
  -- exists. First using find should be more efficient when most strings
  -- don't contain the pattern.
  if strfind (str, pattern) then
    return gsub (str, pattern, repl)
  else
    return str
  end
end

local function quotestring (value)
  -- based on the regexp "escapable" in https://github.com/douglascrockford/JSON-js
  value = fsub (value, "[%z\1-\31\"\\\127]", escapeutf8)
  if strfind (value, "[\194\216\220\225\226\239]") then
    value = fsub (value, "\194[\128-\159\173]", escapeutf8)
    value = fsub (value, "\216[\128-\132]", escapeutf8)
    value = fsub (value, "\220\143", escapeutf8)
    value = fsub (value, "\225\158[\180\181]", escapeutf8)
    value = fsub (value, "\226\128[\140-\143\168-\175]", escapeutf8)
    value = fsub (value, "\226\129[\160-\175]", escapeutf8)
    value = fsub (value, "\239\187\191", escapeutf8)
    value = fsub (value, "\239\191[\176-\191]", escapeutf8)
  end
  return "\"" .. value .. "\""
end
json.quotestring = quotestring

local function replace(str, o, n)
  local i, j = strfind (str, o, 1, true)
  if i then
    return strsub(str, 1, i-1) .. n .. strsub(str, j+1, -1)
  else
    return str
  end
end

-- locale independent num2str and str2num functions
local decpoint, numfilter

local function updatedecpoint ()
  decpoint = strmatch(tostring(0.5), "([^05+])")
  -- build a filter that can be used to remove group separators
  numfilter = "[^0-9%-%+eE" .. gsub(decpoint, "[%^%$%(%)%%%.%[%]%*%+%-%?]", "%%%0") .. "]+"
end

updatedecpoint()

local function num2str (num)
  return replace(fsub(tostring(num), numfilter, ""), decpoint, ".")
end

local function str2num (str)
  local num = tonumber(replace(str, ".", decpoint))
  if not num then
    updatedecpoint()
    num = tonumber(replace(str, ".", decpoint))
  end
  return num
end

local function addnewline2 (level, buffer, buflen)
  buffer[buflen+1] = "\n"
  buffer[buflen+2] = strrep ("  ", level)
  buflen = buflen + 2
  return buflen
end

function json.addnewline (state)
  if state.indent then
    state.bufferlen = addnewline2 (state.level or 0,
                           state.buffer, state.bufferlen or #(state.buffer))
  end
end

local encode2 -- forward declaration

local function addpair (key, value, prev, indent, level, buffer, buflen, tables, globalorder, state)
  local kt = type (key)
  if kt ~= 'string' and kt ~= 'number' then
    return nil, "type '" .. kt .. "' is not supported as a key by JSON."
  end
  if prev then
    buflen = buflen + 1
    buffer[buflen] = ","
  end
  if indent then
    buflen = addnewline2 (level, buffer, buflen)
  end
  buffer[buflen+1] = quotestring (key)
  buffer[buflen+2] = ":"
  return encode2 (value, indent, level, buffer, buflen + 2, tables, globalorder, state)
end

local function appendcustom(res, buffer, state)
  local buflen = state.bufferlen
  if type (res) == 'string' then
    buflen = buflen + 1
    buffer[buflen] = res
  end
  return buflen
end

local function exception(reason, value, state, buffer, buflen, defaultmessage)
  defaultmessage = defaultmessage or reason
  local handler = state.exception
  if not handler then
    return nil, defaultmessage
  else
    state.bufferlen = buflen
    local ret, msg = handler (reason, value, state, defaultmessage)
    if not ret then return nil, msg or defaultmessage end
    return appendcustom(ret, buffer, state)
  end
end

function json.encodeexception(reason, value, state, defaultmessage)
  return quotestring("<" .. defaultmessage .. ">")
end

encode2 = function (value, indent, level, buffer, buflen, tables, globalorder, state)
  local valtype = type (value)
  local valmeta = getmetatable (value)
  valmeta = type (valmeta) == 'table' and valmeta -- only tables
  local valtojson = valmeta and valmeta.__tojson
  if valtojson then
    if tables[value] then
      return exception('reference cycle', value, state, buffer, buflen)
    end
    tables[value] = true
    state.bufferlen = buflen
    local ret, msg = valtojson (value, state)
    if not ret then return exception('custom encoder failed', value, state, buffer, buflen, msg) end
    tables[value] = nil
    buflen = appendcustom(ret, buffer, state)
  elseif value == nil then
    buflen = buflen + 1
    buffer[buflen] = "null"
  elseif valtype == 'number' then
    local s
    if value ~= value or value >= huge or -value >= huge then
      -- This is the behaviour of the original JSON implementation.
      s = "null"
    else
      s = num2str (value)
    end
    buflen = buflen + 1
    buffer[buflen] = s
  elseif valtype == 'boolean' then
    buflen = buflen + 1
    buffer[buflen] = value and "true" or "false"
  elseif valtype == 'string' then
    buflen = buflen + 1
    buffer[buflen] = quotestring (value)
  elseif valtype == 'table' then
    if tables[value] then
      return exception('reference cycle', value, state, buffer, buflen)
    end
    tables[value] = true
    level = level + 1
    local isa, n = isarray (value)
    if n == 0 and valmeta and valmeta.__jsontype == 'object' then
      isa = false
    end
    local msg
    if isa then -- JSON array
      buflen = buflen + 1
      buffer[buflen] = "["
      for i = 1, n do
        buflen, msg = encode2 (value[i], indent, level, buffer, buflen, tables, globalorder, state)
        if not buflen then return nil, msg end
        if i < n then
          buflen = buflen + 1
          buffer[buflen] = ","
        end
      end
      buflen = buflen + 1
      buffer[buflen] = "]"
    else -- JSON object
      local prev = false
      buflen = buflen + 1
      buffer[buflen] = "{"
      local order = valmeta and valmeta.__jsonorder or globalorder
      if order then
        local used = {}
        n = #order
        for i = 1, n do
          local k = order[i]
          local v = value[k]
          if v ~= nil then
            used[k] = true
            buflen, msg = addpair (k, v, prev, indent, level, buffer, buflen, tables, globalorder, state)
            prev = true -- add a seperator before the next element
          end
        end
        for k,v in pairs (value) do
          if not used[k] then
            buflen, msg = addpair (k, v, prev, indent, level, buffer, buflen, tables, globalorder, state)
            if not buflen then return nil, msg end
            prev = true -- add a seperator before the next element
          end
        end
      else -- unordered
        for k,v in pairs (value) do
          buflen, msg = addpair (k, v, prev, indent, level, buffer, buflen, tables, globalorder, state)
          if not buflen then return nil, msg end
          prev = true -- add a seperator before the next element
        end
      end
      if indent then
        buflen = addnewline2 (level - 1, buffer, buflen)
      end
      buflen = buflen + 1
      buffer[buflen] = "}"
    end
    tables[value] = nil
  else
    return exception ('unsupported type', value, state, buffer, buflen,
      "type '" .. valtype .. "' is not supported by JSON.")
  end
  return buflen
end

function json.encode (value, state)
  state = state or {}
  local oldbuffer = state.buffer
  local buffer = oldbuffer or {}
  state.buffer = buffer
  updatedecpoint()
  local ret, msg = encode2 (value, state.indent, state.level or 0,
                   buffer, state.bufferlen or 0, state.tables or {}, state.keyorder, state)
  if not ret then
    error (msg, 2)
  elseif oldbuffer == buffer then
    state.bufferlen = ret
    return true
  else
    state.bufferlen = nil
    state.buffer = nil
    return concat (buffer)
  end
end

local function loc (str, where)
  local line, pos, linepos = 1, 1, 0
  while true do
    pos = strfind (str, "\n", pos, true)
    if pos and pos < where then
      line = line + 1
      linepos = pos
      pos = pos + 1
    else
      break
    end
  end
  return "line " .. line .. ", column " .. (where - linepos)
end

local function unterminated (str, what, where)
  return nil, strlen (str) + 1, "unterminated " .. what .. " at " .. loc (str, where)
end

local function scanwhite (str, pos)
  while true do
    pos = strfind (str, "%S", pos)
    if not pos then return nil end
    local sub2 = strsub (str, pos, pos + 1)
    if sub2 == "\239\187" and strsub (str, pos + 2, pos + 2) == "\191" then
      -- UTF-8 Byte Order Mark
      pos = pos + 3
    elseif sub2 == "//" then
      pos = strfind (str, "[\n\r]", pos + 2)
      if not pos then return nil end
    elseif sub2 == "/*" then
      pos = strfind (str, "*/", pos + 2)
      if not pos then return nil end
      pos = pos + 2
    else
      return pos
    end
  end
end

local escapechars = {
  ["\""] = "\"", ["\\"] = "\\", ["/"] = "/", ["b"] = "\b", ["f"] = "\f",
  ["n"] = "\n", ["r"] = "\r", ["t"] = "\t"
}

local function unichar (value)
  if value < 0 then
    return nil
  elseif value <= 0x007f then
    return strchar (value)
  elseif value <= 0x07ff then
    return strchar (0xc0 + floor(value/0x40),
                    0x80 + (floor(value) % 0x40))
  elseif value <= 0xffff then
    return strchar (0xe0 + floor(value/0x1000),
                    0x80 + (floor(value/0x40) % 0x40),
                    0x80 + (floor(value) % 0x40))
  elseif value <= 0x10ffff then
    return strchar (0xf0 + floor(value/0x40000),
                    0x80 + (floor(value/0x1000) % 0x40),
                    0x80 + (floor(value/0x40) % 0x40),
                    0x80 + (floor(value) % 0x40))
  else
    return nil
  end
end

local function scanstring (str, pos)
  local lastpos = pos + 1
  local buffer, n = {}, 0
  while true do
    local nextpos = strfind (str, "[\"\\]", lastpos)
    if not nextpos then
      return unterminated (str, "string", pos)
    end
    if nextpos > lastpos then
      n = n + 1
      buffer[n] = strsub (str, lastpos, nextpos - 1)
    end
    if strsub (str, nextpos, nextpos) == "\"" then
      lastpos = nextpos + 1
      break
    else
      local escchar = strsub (str, nextpos + 1, nextpos + 1)
      local value
      if escchar == "u" then
        value = tonumber (strsub (str, nextpos + 2, nextpos + 5), 16)
        if value then
          local value2
          if 0xD800 <= value and value <= 0xDBff then
            -- we have the high surrogate of UTF-16. Check if there is a
            -- low surrogate escaped nearby to combine them.
            if strsub (str, nextpos + 6, nextpos + 7) == "\\u" then
              value2 = tonumber (strsub (str, nextpos + 8, nextpos + 11), 16)
              if value2 and 0xDC00 <= value2 and value2 <= 0xDFFF then
                value = (value - 0xD800)  * 0x400 + (value2 - 0xDC00) + 0x10000
              else
                value2 = nil -- in case it was out of range for a low surrogate
              end
            end
          end
          value = value and unichar (value)
          if value then
            if value2 then
              lastpos = nextpos + 12
            else
              lastpos = nextpos + 6
            end
          end
        end
      end
      if not value then
        value = escapechars[escchar] or escchar
        lastpos = nextpos + 2
      end
      n = n + 1
      buffer[n] = value
    end
  end
  if n == 1 then
    return buffer[1], lastpos
  elseif n > 1 then
    return concat (buffer), lastpos
  else
    return "", lastpos
  end
end

local scanvalue -- forward declaration

local function scantable (what, closechar, str, startpos, nullval, objectmeta, arraymeta)
  local len = strlen (str)
  local tbl, n = {}, 0
  local pos = startpos + 1
  if what == 'object' then
    setmetatable (tbl, objectmeta)
  else
    setmetatable (tbl, arraymeta)
  end
  while true do
    pos = scanwhite (str, pos)
    if not pos then return unterminated (str, what, startpos) end
    local char = strsub (str, pos, pos)
    if char == closechar then
      return tbl, pos + 1
    end
    local val1, err
    val1, pos, err = scanvalue (str, pos, nullval, objectmeta, arraymeta)
    if err then return nil, pos, err end
    pos = scanwhite (str, pos)
    if not pos then return unterminated (str, what, startpos) end
    char = strsub (str, pos, pos)
    if char == ":" then
      if val1 == nil then
        return nil, pos, "cannot use nil as table index (at " .. loc (str, pos) .. ")"
      end
      pos = scanwhite (str, pos + 1)
      if not pos then return unterminated (str, what, startpos) end
      local val2
      val2, pos, err = scanvalue (str, pos, nullval, objectmeta, arraymeta)
      if err then return nil, pos, err end
      tbl[val1] = val2
      pos = scanwhite (str, pos)
      if not pos then return unterminated (str, what, startpos) end
      char = strsub (str, pos, pos)
    else
      n = n + 1
      tbl[n] = val1
    end
    if char == "," then
      pos = pos + 1
    end
  end
end

scanvalue = function (str, pos, nullval, objectmeta, arraymeta)
  pos = pos or 1
  pos = scanwhite (str, pos)
  if not pos then
    return nil, strlen (str) + 1, "no valid JSON value (reached the end)"
  end
  local char = strsub (str, pos, pos)
  if char == "{" then
    return scantable ('object', "}", str, pos, nullval, objectmeta, arraymeta)
  elseif char == "[" then
    return scantable ('array', "]", str, pos, nullval, objectmeta, arraymeta)
  elseif char == "\"" then
    return scanstring (str, pos)
  else
    local pstart, pend = strfind (str, "^%-?[%d%.]+[eE]?[%+%-]?%d*", pos)
    if pstart then
      local number = str2num (strsub (str, pstart, pend))
      if number then
        return number, pend + 1
      end
    end
    pstart, pend = strfind (str, "^%a%w*", pos)
    if pstart then
      local name = strsub (str, pstart, pend)
      if name == "true" then
        return true, pend + 1
      elseif name == "false" then
        return false, pend + 1
      elseif name == "null" then
        return nullval, pend + 1
      end
    end
    return nil, pos, "no valid JSON value at " .. loc (str, pos)
  end
end

local function optionalmetatables(...)
  if select("#", ...) > 0 then
    return ...
  else
    return {__jsontype = 'object'}, {__jsontype = 'array'}
  end
end

function json.decode (str, pos, nullval, ...)
  local objectmeta, arraymeta = optionalmetatables(...)
  return scanvalue (str, pos, nullval, objectmeta, arraymeta)
end

function json.use_lpeg ()
  local g = require ("lpeg")

  if g.version() == "0.11" then
    error "due to a bug in LPeg 0.11, it cannot be used for JSON matching"
  end

  local pegmatch = g.match
  local P, S, R = g.P, g.S, g.R

  local function ErrorCall (str, pos, msg, state)
    if not state.msg then
      state.msg = msg .. " at " .. loc (str, pos)
      state.pos = pos
    end
    return false
  end

  local function Err (msg)
    return g.Cmt (g.Cc (msg) * g.Carg (2), ErrorCall)
  end

  local function ErrorUnterminatedCall (str, pos, what, state)
    return ErrorCall (str, pos - 1, "unterminated " .. what, state)
  end

  local SingleLineComment = P"//" * (1 - S"\n\r")^0
  local MultiLineComment = P"/*" * (1 - P"*/")^0 * P"*/"
  local Space = (S" \n\r\t" + P"\239\187\191" + SingleLineComment + MultiLineComment)^0

  local function ErrUnterminated (what)
    return g.Cmt (g.Cc (what) * g.Carg (2), ErrorUnterminatedCall)
  end

  local PlainChar = 1 - S"\"\\\n\r"
  local EscapeSequence = (P"\\" * g.C (S"\"\\/bfnrt" + Err "unsupported escape sequence")) / escapechars
  local HexDigit = R("09", "af", "AF")
  local function UTF16Surrogate (match, pos, high, low)
    high, low = tonumber (high, 16), tonumber (low, 16)
    if 0xD800 <= high and high <= 0xDBff and 0xDC00 <= low and low <= 0xDFFF then
      return true, unichar ((high - 0xD800)  * 0x400 + (low - 0xDC00) + 0x10000)
    else
      return false
    end
  end
  local function UTF16BMP (hex)
    return unichar (tonumber (hex, 16))
  end
  local U16Sequence = (P"\\u" * g.C (HexDigit * HexDigit * HexDigit * HexDigit))
  local UnicodeEscape = g.Cmt (U16Sequence * U16Sequence, UTF16Surrogate) + U16Sequence/UTF16BMP
  local Char = UnicodeEscape + EscapeSequence + PlainChar
  local String = P"\"" * (g.Cs (Char ^ 0) * P"\"" + ErrUnterminated "string")
  local Integer = P"-"^(-1) * (P"0" + (R"19" * R"09"^0))
  local Fractal = P"." * R"09"^0
  local Exponent = (S"eE") * (S"+-")^(-1) * R"09"^1
  local Number = (Integer * Fractal^(-1) * Exponent^(-1))/str2num
  local Constant = P"true" * g.Cc (true) + P"false" * g.Cc (false) + P"null" * g.Carg (1)
  local SimpleValue = Number + String + Constant
  local ArrayContent, ObjectContent

  -- The functions parsearray and parseobject parse only a single value/pair
  -- at a time and store them directly to avoid hitting the LPeg limits.
  local function parsearray (str, pos, nullval, state)
    local obj, cont
    local start = pos
    local npos
    local t, nt = {}, 0
    repeat
      obj, cont, npos = pegmatch (ArrayContent, str, pos, nullval, state)
      if cont == 'end' then
        return ErrorUnterminatedCall (str, start, "array", state)
      end
      pos = npos
      if cont == 'cont' or cont == 'last' then
        nt = nt + 1
        t[nt] = obj
      end
    until cont ~= 'cont'
    return pos, setmetatable (t, state.arraymeta)
  end

  local function parseobject (str, pos, nullval, state)
    local obj, key, cont
    local start = pos
    local npos
    local t = {}
    repeat
      key, obj, cont, npos = pegmatch (ObjectContent, str, pos, nullval, state)
      if cont == 'end' then
        return ErrorUnterminatedCall (str, start, "object", state)
      end
      pos = npos
      if cont == 'cont' or cont == 'last' then
        t[key] = obj
      end
    until cont ~= 'cont'
    return pos, setmetatable (t, state.objectmeta)
  end

  local Array = P"[" * g.Cmt (g.Carg(1) * g.Carg(2), parsearray)
  local Object = P"{" * g.Cmt (g.Carg(1) * g.Carg(2), parseobject)
  local Value = Space * (Array + Object + SimpleValue)
  local ExpectedValue = Value + Space * Err "value expected"
  local ExpectedKey = String + Err "key expected"
  local End = P(-1) * g.Cc'end'
  local ErrInvalid = Err "invalid JSON"
  ArrayContent = (Value * Space * (P"," * g.Cc'cont' + P"]" * g.Cc'last'+ End + ErrInvalid)  + g.Cc(nil) * (P"]" * g.Cc'empty' + End  + ErrInvalid)) * g.Cp()
  local Pair = g.Cg (Space * ExpectedKey * Space * (P":" + Err "colon expected") * ExpectedValue)
  ObjectContent = (g.Cc(nil) * g.Cc(nil) * P"}" * g.Cc'empty' + End + (Pair * Space * (P"," * g.Cc'cont' + P"}" * g.Cc'last' + End + ErrInvalid) + ErrInvalid)) * g.Cp()
  local DecodeValue = ExpectedValue * g.Cp ()

  jsonlpeg.version = json.version
  jsonlpeg.encode = json.encode
  jsonlpeg.null = json.null
  jsonlpeg.quotestring = json.quotestring
  jsonlpeg.addnewline = json.addnewline
  jsonlpeg.encodeexception = json.encodeexception
  jsonlpeg.using_lpeg = true

  function jsonlpeg.decode (str, pos, nullval, ...)
    local state = {}
    state.objectmeta, state.arraymeta = optionalmetatables(...)
    local obj, retpos = pegmatch (DecodeValue, str, pos, nullval, state)
    if state.msg then
      return nil, state.pos, state.msg
    else
      return obj, retpos
    end
  end

  -- cache result of this function:
  json.use_lpeg = function () return jsonlpeg end
  jsonlpeg.use_lpeg = json.use_lpeg

  return jsonlpeg
end

if always_use_lpeg then
  return json.use_lpeg()
end

return json
end
end

do
local _ENV = _ENV
package.preload[ "modules.ljsocket" ] = function( ... ) local arg = _G.arg;
local ffi = require("ffi")
local socket = {}
local e = {}
local errno = {}

local assert = assert;
local bit = bit;
local jit = jit;
local string = string;
local tostring = tostring;
local pairs = pairs;
local ipairs = ipairs;
local error = error;
local type = type;
local tonumber = tonumber;
local table = table;
local setmetatable = setmetatable;
local unpack = unpack;
local print = print;

do
    local C

    if ffi.os == "Windows" then
        C = assert(ffi.load("ws2_32"))
    else
        C = ffi.C
    end

    local M = {}

    local function generic_function(C_name, cdef, alias, size_error_handling)
        ffi.cdef(cdef)

        alias = alias or C_name
        local func_name = alias
        local func = C[C_name]

        if size_error_handling == false then
            socket[func_name] = func
        elseif size_error_handling then
            socket[func_name] = function(...)
                local len = func(...)
                if len < 0 then
                    return nil, socket.lasterror()
                end

                return len
            end
        else
            socket[func_name] = function(...)
                local ret = func(...)

                if ret == 0 then
                    return true
                end

                return nil, socket.lasterror()
            end
        end
    end

    ffi.cdef[[
        struct in_addr {
            uint32_t s_addr;
        };

        struct in6_addr {
            union {
                uint8_t u6_addr8[16];
                uint16_t u6_addr16[8];
                uint32_t u6_addr32[4];
            } u6_addr;
        };
    ]]

    -- https://www.cs.dartmouth.edu/~sergey/cs60/on-sockaddr-structs.txt

    if ffi.os == "OSX" then
        ffi.cdef[[
            struct sockaddr {
                uint8_t sa_len;
                uint8_t sa_family;
                char sa_data[14];
            };

            struct sockaddr_in {
                uint8_t sin_len;
                uint8_t sin_family;
                uint16_t sin_port;
                struct in_addr sin_addr;
                char sin_zero[8];
            };

            struct sockaddr_in6 {
                uint8_t sin6_len;
                uint8_t sin6_family;
                uint16_t sin6_port;
                uint32_t sin6_flowinfo;
                struct in6_addr sin6_addr;
                uint32_t sin6_scope_id;
            };
        ]]
    elseif ffi.os == "Windows" then
        ffi.cdef[[
            struct sockaddr {
                uint16_t sa_family;
                char sa_data[14];
            };

            struct sockaddr_in {
                int16_t sin_family;
                uint16_t sin_port;
                struct in_addr sin_addr;
                uint8_t sin_zero[8];
            };

            struct sockaddr_in6 {
                int16_t sin6_family;
                uint16_t sin6_port;
                uint32_t sin6_flowinfo;
                struct in6_addr sin6_addr;
                uint32_t sin6_scope_id;
            };
        ]]
    else -- posix
        ffi.cdef[[
            struct sockaddr {
                uint16_t sa_family;
                char sa_data[14];
            };

            struct sockaddr_in {
                uint16_t sin_family;
                uint16_t sin_port;
                struct in_addr sin_addr;
                char sin_zero[8];
            };

            struct sockaddr_in6 {
                uint16_t sin6_family;
                uint16_t sin6_port;
                uint32_t sin6_flowinfo;
                struct in6_addr sin6_addr;
                uint32_t sin6_scope_id;
            };
        ]]
    end

    if ffi.os == "Windows" then
        ffi.cdef[[
            typedef size_t SOCKET;

            struct addrinfo {
                int ai_flags;
                int ai_family;
                int ai_socktype;
                int ai_protocol;
                size_t ai_addrlen;
                char *ai_canonname;
                struct sockaddr *ai_addr;
                struct addrinfo *ai_next;
            };
        ]]
        socket.INVALID_SOCKET = ffi.new("SOCKET", -1)
    elseif ffi.os == "OSX" then
        ffi.cdef[[
            typedef int32_t SOCKET;

            struct addrinfo {
                int ai_flags;
                int ai_family;
                int ai_socktype;
                int ai_protocol;
                uint32_t ai_addrlen;
                char *ai_canonname;
                struct sockaddr *ai_addr;
                struct addrinfo *ai_next;
            };
        ]]
        socket.INVALID_SOCKET = -1
    else
        ffi.cdef[[
            typedef int32_t SOCKET;

            struct addrinfo {
                int ai_flags;
                int ai_family;
                int ai_socktype;
                int ai_protocol;
                uint32_t ai_addrlen;
                struct sockaddr *ai_addr;
                char *ai_canonname;
                struct addrinfo *ai_next;
            };
        ]]
        socket.INVALID_SOCKET = -1
    end

    assert(ffi.sizeof("struct sockaddr") == 16)
    assert(ffi.sizeof("struct sockaddr_in") == 16)

    if ffi.os == "Windows" then
        ffi.cdef[[

            struct pollfd {
                SOCKET fd;
                short events;
                short revents;
            };
            int WSAPoll(struct pollfd *fds, unsigned long int nfds, int timeout);

            uint32_t GetLastError();
            uint32_t FormatMessageA(
                uint32_t dwFlags,
                const void* lpSource,
                uint32_t dwMessageId,
                uint32_t dwLanguageId,
                char* lpBuffer,
                uint32_t nSize,
                va_list *Arguments
            );
        ]]

        local function WORD(low, high)
            return bit.bor(low , bit.lshift(high , 8))
        end

        do
            ffi.cdef[[int GetLastError();]]

            local FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000
            local FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200
            local flags = bit.bor(FORMAT_MESSAGE_IGNORE_INSERTS, FORMAT_MESSAGE_FROM_SYSTEM)

            local cache = {}

            function socket.lasterror(num)
                num = num or ffi.C.GetLastError()

                if not cache[num] then
                    local buffer = ffi.new("char[512]")
                    local len = ffi.C.FormatMessageA(flags, nil, num, 0, buffer, ffi.sizeof(buffer), nil)
                    cache[num] = ffi.string(buffer, len - 2)
                end

                return cache[num], num
            end
        end

        do
            ffi.cdef[[int WSAStartup(uint16_t version, void *wsa_data);]]

            local wsa_data

            if jit.arch == "x64" then
                wsa_data = ffi.typeof([[struct {
                    uint16_t wVersion;
                    uint16_t wHighVersion;
                    unsigned short iMax_M;
                    unsigned short iMaxUdpDg;
                    char * lpVendorInfo;
                    char szDescription[257];
                    char szSystemStatus[129];
                }]])
            else
                wsa_data = ffi.typeof([[struct {
                    uint16_t wVersion;
                    uint16_t wHighVersion;
                    char szDescription[257];
                    char szSystemStatus[129];
                    unsigned short iMax_M;
                    unsigned short iMaxUdpDg;
                    char * lpVendorInfo;
                }]])
            end

            function socket.initialize()
                local data = wsa_data()

                if C.WSAStartup(WORD(2, 2), data) == 0 then
                    return data
                end

                return nil, socket.lasterror()
            end
        end

        do
            ffi.cdef[[int WSACleanup();]]

            function socket.shutdown()
                if C.WSACleanup() == 0 then
                    return true
                end

                return nil, socket.lasterror()
            end
        end

        if jit.arch ~= "x64" then -- xp or something
            ffi.cdef[[int WSAAddressToStringA(struct sockaddr *, unsigned long, void *, char *, unsigned long *);]]

            function socket.inet_ntop(family, pAddr, strptr, strlen)
                -- win XP: http://memset.wordpress.com/2010/10/09/inet_ntop-for-win32/
                local srcaddr = ffi.new("struct sockaddr_in")
                ffi.copy(srcaddr.sin_addr, pAddr, ffi.sizeof(srcaddr.sin_addr))
                srcaddr.sin_family = family
                local len = ffi.new("unsigned long[1]", strlen)
                C.WSAAddressToStringA(ffi.cast("struct sockaddr *", srcaddr), ffi.sizeof(srcaddr), nil, strptr, len)
                return strptr
            end
        end

        generic_function("closesocket", "int closesocket(SOCKET s);", "close")

        do
            ffi.cdef[[int ioctlsocket(SOCKET s, long cmd, unsigned long* argp);]]

            local IOCPARM_MASK    = 0x7
            local IOC_IN          = 0x80000000
            local function _IOW(x,y,t)
                return bit.bor(IOC_IN, bit.lshift(bit.band(ffi.sizeof(t),IOCPARM_MASK),16), bit.lshift(x,8), y)
            end

            local FIONBIO = _IOW(string.byte'f', 126, "uint32_t") -- -2147195266 -- 2147772030ULL

            function socket.blocking(fd, b)
                local ret = C.ioctlsocket(fd, FIONBIO, ffi.new("int[1]", b and 0 or 1))
                if ret == 0 then
                    return true
                end

                return nil, socket.lasterror()
            end
        end

        function socket.poll(fds, ndfs, timeout)
            local ret = C.WSAPoll(fds, ndfs, timeout)
            if ret < 0 then
                return nil, socket.lasterror()
            end
            return ret
        end
    else
        ffi.cdef[[
            struct pollfd {
                SOCKET fd;
                short events;
                short revents;
            };

            int poll(struct pollfd *fds, unsigned long nfds, int timeout);
        ]]

        do
            local cache = {}

            function socket.lasterror(num)
                num = num or ffi.errno()

                if not cache[num] then
                    local err = ffi.string(ffi.C.strerror(num))
                    cache[num] = err == "" and tostring(num) or err
                end

                return cache[num], num
            end
        end

        generic_function("close", "int close(SOCKET s);")

        do
            ffi.cdef[[int fcntl(int, int, ...);]]

            local F_GETFL = 3
            local F_SETFL = 4
            local O_NONBLOCK = 04000

            if ffi.os == "OSX" then
                O_NONBLOCK = 0x0004
            end
            function socket.blocking(fd, b)
                local flags = ffi.C.fcntl(fd, F_GETFL, 0)

                if flags < 0 then
                    -- error
                    return nil, socket.lasterror()
                end

                if b then
                    flags = bit.band(flags, bit.bnot(O_NONBLOCK))
                else
                    flags = bit.bor(flags, O_NONBLOCK)
                end

                local ret = ffi.C.fcntl(fd, F_SETFL, ffi.new("int", flags))

                if ret < 0 then
                    return nil, socket.lasterror()
                end

                return true
            end
        end

        function socket.poll(fds, ndfs, timeout)
            local ret = C.poll(fds, ndfs, timeout)
            if ret < 0 then
                return nil, socket.lasterror()
            end
            return ret
        end
    end


    ffi.cdef[[
        char *strerror(int errnum);
        int getaddrinfo(const char *node, const char *service, const struct addrinfo *hints, struct addrinfo **res);
        int getnameinfo(const struct sockaddr* sa, uint32_t salen, char* host, size_t hostlen, char* serv, size_t servlen, int flags);
        void freeaddrinfo(struct addrinfo *ai);
        const char *gai_strerror(int errcode);
        char *inet_ntoa(struct in_addr in);
        uint16_t ntohs(uint16_t netshort);
    ]]

    function socket.getaddrinfo(node_name, service_name, hints, result)
        local ret = C.getaddrinfo(node_name, service_name, hints, result)
        if ret == 0 then
            return true
        end

        return nil, ffi.string(socket.lasterror(ret))
    end

    function socket.getnameinfo(address, length, host, hostlen, serv, servlen, flags)
        local ret = C.getnameinfo(address, length, host, hostlen, serv, servlen, flags)
        if ret == 0 then
            return true
        end

        return nil, ffi.string(socket.lasterror(ret))
    end

    do
        ffi.cdef[[const char *inet_ntop(int __af, const void *__cp, char *__buf, unsigned int __len);]]

        function socket.inet_ntop(family, addrinfo, strptr, strlen)
            if C.inet_ntop(family, addrinfo, strptr, strlen) == nil then
                return nil, socket.lasterror()
            end

            return strptr
        end
    end

    do
        ffi.cdef[[SOCKET socket(int af, int type, int protocol);]]

        function socket.create(af, type, protocol)
            local fd = C.socket(af, type, protocol)

            if fd <= 0 then
                return nil, socket.lasterror()
            end

            return fd
        end
    end

    generic_function("shutdown", "int shutdown(SOCKET s, int how);")

    generic_function("setsockopt", "int setsockopt(SOCKET s, int level, int optname, const void* optval, uint32_t optlen);")
    generic_function("getsockopt", "int getsockopt(SOCKET s, int level, int optname, void *optval, uint32_t *optlen);")

    generic_function("accept", "SOCKET accept(SOCKET s, struct sockaddr *, int *);", nil, false)
    generic_function("bind", "int bind(SOCKET s, const struct sockaddr* name, int namelen);")
    generic_function("connect", "int connect(SOCKET s, const struct sockaddr * name, int namelen);")

    generic_function("listen", "int listen(SOCKET s, int backlog);")
    generic_function("recv", "int recv(SOCKET s, char* buf, int len, int flags);", nil, true)
    generic_function("recvfrom", "int recvfrom(SOCKET s, char* buf, int len, int flags, struct sockaddr *src_addr, unsigned int *addrlen);", nil, true)

    generic_function("send", "int send(SOCKET s, const char* buf, int len, int flags);", nil, true)
    generic_function("sendto", "int sendto(SOCKET s, const char* buf, int len, int flags, const struct sockaddr* to, int tolen);", nil, true)

    generic_function("getpeername", "int getpeername(SOCKET s, struct sockaddr *, unsigned int *);")
    generic_function("getsockname", "int getsockname(SOCKET s, struct sockaddr *, unsigned int *);")

    socket.inet_ntoa = C.inet_ntoa
    socket.ntohs = C.ntohs

    function socket.poll(fd, events, revents)

    end

    e = {
        TCP_NODELAY = 1,
        TCP_MAXSEG = 2,
        TCP_CORK = 3,
        TCP_KEEPIDLE = 4,
        TCP_KEEPINTVL = 5,
        TCP_KEEPCNT = 6,
        TCP_SYNCNT = 7,
        TCP_LINGER2 = 8,
        TCP_DEFER_ACCEPT = 9,
        TCP_WINDOW_CLAMP = 10,
        TCP_INFO = 11,
        TCP_QUICKACK = 12,
        TCP_CONGESTION = 13,
        TCP_MD5SIG = 14,
        TCP_THIN_LINEAR_TIMEOUTS = 16,
        TCP_THIN_DUPACK = 17,
        TCP_USER_TIMEOUT = 18,
        TCP_REPAIR = 19,
        TCP_REPAIR_QUEUE = 20,
        TCP_QUEUE_SEQ = 21,
        TCP_REPAIR_OPTIONS = 22,
        TCP_FASTOPEN = 23,
        TCP_TIMESTAMP = 24,
        TCP_NOTSENT_LOWAT = 25,
        TCP_CC_INFO = 26,
        TCP_SAVE_SYN = 27,
        TCP_SAVED_SYN = 28,
        TCP_REPAIR_WINDOW = 29,
        TCP_FASTOPEN_CONNECT = 30,
        TCP_ULP = 31,
        TCP_MD5SIG_EXT = 32,
        TCP_FASTOPEN_KEY = 33,
        TCP_FASTOPEN_NO_COOKIE = 34,
        TCP_ZEROCOPY_RECEIVE = 35,
        TCP_INQ = 36,

        AF_INET = 2,
        AF_INET6 = 10,
        AF_UNSPEC = 0,

        AF_UNIX = 1,
        AF_AX25 = 3,
        AF_IPX = 4,
        AF_APPLETALK = 5,
        AF_NETROM = 6,
        AF_BRIDGE = 7,
        AF_AAL5 = 8,
        AF_X25 = 9,

        INET6_ADDRSTRLEN = 46,
        INET_ADDRSTRLEN = 16,

        SO_DEBUG = 1,
        SO_REUSEADDR = 2,
        SO_TYPE = 3,
        SO_ERROR = 4,
        SO_DONTROUTE = 5,
        SO_BROADCAST = 6,
        SO_SNDBUF = 7,
        SO_RCVBUF = 8,
        SO_SNDBUFFORCE = 32,
        SO_RCVBUFFORCE = 33,
        SO_KEEPALIVE = 9,
        SO_OOBINLINE = 10,
        SO_NO_CHECK = 11,
        SO_PRIORITY = 12,
        SO_LINGER = 13,
        SO_BSDCOMPAT = 14,
        SO_REUSEPORT = 15,
        SO_PASSCRED = 16,
        SO_PEERCRED = 17,
        SO_RCVLOWAT = 18,
        SO_SNDLOWAT = 19,
        SO_RCVTIMEO = 20,
        SO_SNDTIMEO = 21,
        SO_SECURITY_AUTHENTICATION = 22,
        SO_SECURITY_ENCRYPTION_TRANSPORT = 23,
        SO_SECURITY_ENCRYPTION_NETWORK = 24,
        SO_BINDTODEVICE = 25,
        SO_ATTACH_FILTER = 26,
        SO_DETACH_FILTER = 27,
        SO_GET_FILTER = 26,
        SO_PEERNAME = 28,
        SO_TIMESTAMP = 29,
        SO_ACCEPTCONN = 30,
        SO_PEERSEC = 31,
        SO_PASSSEC = 34,
        SO_TIMESTAMPNS = 35,
        SO_MARK = 36,
        SO_TIMESTAMPING = 37,
        SO_PROTOCOL = 38,
        SO_DOMAIN = 39,
        SO_RXQ_OVFL = 40,
        SO_WIFI_STATUS = 41,
        SO_PEEK_OFF = 42,
        SO_NOFCS = 43,
        SO_LOCK_FILTER = 44,
        SO_SELECT_ERR_QUEUE = 45,
        SO_BUSY_POLL = 46,
        SO_MAX_PACING_RATE = 47,
        SO_BPF_EXTENSIONS = 48,
        SO_INCOMING_CPU = 49,
        SO_ATTACH_BPF = 50,
        SO_DETACH_BPF = 27,
        SO_ATTACH_REUSEPORT_CBPF = 51,
        SO_ATTACH_REUSEPORT_EBPF = 52,
        SO_CNX_ADVICE = 53,
        SO_MEMINFO = 55,
        SO_INCOMING_NAPI_ID = 56,
        SO_COOKIE = 57,
        SO_PEERGROUPS = 59,
        SO_ZEROCOPY = 60,
        SO_TXTIME = 61,
        SOL_SOCKET = 1,
        SOL_TCP = 6,

        SOMAXCONN = 128,

        IPPROTO_IP = 0,
        IPPROTO_HOPOPTS = 0,
        IPPROTO_ICMP = 1,
        IPPROTO_IGMP = 2,
        IPPROTO_IPIP = 4,
        IPPROTO_TCP = 6,
        IPPROTO_EGP = 8,
        IPPROTO_PUP = 12,
        IPPROTO_UDP = 17,
        IPPROTO_IDP = 22,
        IPPROTO_TP = 29,
        IPPROTO_DCCP = 33,
        IPPROTO_IPV6 = 41,
        IPPROTO_ROUTING = 43,
        IPPROTO_FRAGMENT = 44,
        IPPROTO_RSVP = 46,
        IPPROTO_GRE = 47,
        IPPROTO_ESP = 50,
        IPPROTO_AH = 51,
        IPPROTO_ICMPV6 = 58,
        IPPROTO_NONE = 59,
        IPPROTO_DSTOPTS = 60,
        IPPROTO_MTP = 92,
        IPPROTO_ENCAP = 98,
        IPPROTO_PIM = 103,
        IPPROTO_COMP = 108,
        IPPROTO_SCTP = 132,
        IPPROTO_UDPLITE = 136,
        IPPROTO_RAW = 255,

        SOCK_STREAM = 1,
        SOCK_DGRAM = 2,
        SOCK_RAW = 3,
        SOCK_RDM = 4,
        SOCK_SEQPACKET = 5,
        SOCK_DCCP = 6,
        SOCK_PACKET = 10,
        SOCK_CLOEXEC = 02000000,
        SOCK_NONBLOCK = 04000,

        AI_PASSIVE = 0x00000001,
        AI_CANONNAME = 0x00000002,
        AI_NUMERICHOST = 0x00000004,
        AI_NUMERICSERV = 0x00000008,
        AI_ALL = 0x00000100,
        AI_ADDRCONFIG = 0x00000400,
        AI_V4MAPPED = 0x00000800,
        AI_NON_AUTHORITATIVE = 0x00004000,
        AI_SECURE = 0x00008000,
        AI_RETURN_PREFERRED_NAMES = 0x00010000,
        AI_FQDN = 0x00020000,
        AI_FILESERVER = 0x00040000,

        POLLIN = 0x0001,
        POLLPRI = 0x0002,
        POLLOUT = 0x0004,
        POLLRDNORM = 0x0040,
        POLLWRNORM = 0x0004,
        POLLRDBAND = 0x0080,
        POLLWRBAND = 0x0100,
        POLLEXTEND = 0x0200,
        POLLATTRIB = 0x0400,
        POLLNLINK = 0x0800,
        POLLWRITE = 0x1000,
        POLLERR = 0x0008,
        POLLHUP = 0x0010,
        POLLNVAL = 0x0020,

        MSG_OOB = 0x01,
        MSG_PEEK = 0x02,
        MSG_DONTROUTE = 0x04,
        MSG_CTRUNC = 0x08,
        MSG_PROXY = 0x10,
        MSG_TRUNC = 0x20,
        MSG_DONTWAIT = 0x40,
        MSG_EOR = 0x80,
        MSG_WAITALL = 0x100,
        MSG_FIN = 0x200,
        MSG_SYN = 0x400,
        MSG_CONFIRM = 0x800,
        MSG_RST = 0x1000,
        MSG_ERRQUEUE = 0x2000,
        MSG_NOSIGNAL = 0x4000,
        MSG_MORE = 0x8000,
        MSG_WAITFORONE = 0x10000,
        MSG_CMSG_CLOEXEC = 0x40000000,
    }

    errno = {
        EAGAIN = 11,
        EWOULDBLOCK = 11, -- is errno.EAGAIN
        EINVAL = 22,
        ENOTSOCK = 88,
        ECONNRESET = 104,
        EINPROGRESS = 115,
    }

    if ffi.os == "Windows" then
        e.SO_SNDLOWAT = 4099
        e.SO_REUSEADDR = 4
        e.SO_KEEPALIVE = 8
        e.SOMAXCONN = 2147483647
        e.AF_INET6 = 23
        e.SO_RCVTIMEO = 4102
        e.SOL_SOCKET = 65535
        e.SO_LINGER = 128
        e.SO_OOBINLINE = 256
        e.POLLWRNORM = 16
        e.SO_ERROR = 4103
        e.SO_BROADCAST = 32
        e.SO_ACCEPTCONN = 2
        e.SO_RCVBUF = 4098
        e.SO_SNDTIMEO = 4101
        e.POLLIN = 768
        e.POLLPRI = 1024
        e.SO_TYPE = 4104
        e.POLLRDBAND = 512
        e.POLLWRBAND = 32
        e.SO_SNDBUF = 4097
        e.POLLNVAL = 4
        e.POLLHUP = 2
        e.POLLERR = 1
        e.POLLRDNORM = 256
        e.SO_DONTROUTE = 16
        e.SO_RCVLOWAT = 4100

        errno.EINVAL = 10022
        errno.EAGAIN = 10035 -- Note: Does not exist on Windows
        errno.EWOULDBLOCK = 10035
        errno.EINPROGRESS = 10036
        errno.ENOTSOCK = 10038
        errno.ECONNRESET = 10054
    end

    if ffi.os == "OSX" then
        e.SOL_SOCKET = 0xffff
        e.SO_DEBUG = 0x0001
        e.SO_ACCEPTCONN = 0x0002
        e.SO_REUSEADDR = 0x0004
        e.SO_KEEPALIVE = 0x0008
        e.SO_DONTROUTE = 0x0010
        e.SO_BROADCAST = 0x0020

        errno.EINVAL = 22
        errno.EAGAIN = 35
        errno.EWOULDBLOCK = errno.EAGAIN
        errno.EINPROGRESS = 36
        errno.ENOTSOCK = 38
        errno.ECONNRESET = 54
    end

    if socket.initialize then
        assert(socket.initialize())
    end
end

local function capture_flags(what)
    local flags = {}
    local reverse = {}
    for k, v in pairs(e) do
        if k:sub(0, #what) == what then
            k = k:sub(#what + 1):lower()
            reverse[v] = k
            flags[k] = v
        end
    end
    return {
        lookup = flags,
        reverse = reverse,
        strict_reverse = function(key)
            if not key then
                error("invalid " .. what:sub(0, -2) .. " flag: nil")
            end
            if not reverse[key] then
                error("invalid "..what:sub(0, -2).." flag: " .. key, 2)
            end
            return reverse[key]
        end,
        strict_lookup = function(key)
            if not key then
                error("invalid " .. what:sub(0, -2) .. " flag: nil")
            end
            if not flags[key] then
                error("invalid "..what:sub(0, -2).." flag: " .. key, 2)
            end
            return flags[key]
        end
    }
end

local SOCK = capture_flags("SOCK_")
local AF = capture_flags("AF_")
local IPPROTO = capture_flags("IPPROTO_")
local AI = capture_flags("AI_")
local SOL = capture_flags("SOL_")
local SO = capture_flags("SO_")
local TCP = capture_flags("TCP_")
local POLL = capture_flags("POLL")

local function table_to_flags(flags, valid_flags, operation)
	if type(flags) == "string" then
		flags = {flags}
    end
    operation = operation or bit.band

	local out = 0

	for k, v in pairs(flags) do
		local flag = valid_flags[v] or valid_flags[k]
		if not flag then
            error("invalid flag " .. tostring(v), 2)
		end

		out = operation(out, tonumber(flag))
	end

	return out
end

local function flags_to_table(flags, valid_flags, operation)
    if not flags then return valid_flags.default_valid_flag end
    operation = operation or bit.band

	local out = {}

	for k, v in pairs(valid_flags) do
		if operation(flags, v) > 0 then
			out[k] = true
		end
	end

	return out
end

local M = {}

local timeout_messages = {}
timeout_messages[errno.EINPROGRESS] = true
timeout_messages[errno.EAGAIN] = true
timeout_messages[errno.EWOULDBLOCK] = true

function M.poll(socket, flags, timeout)
    local pfd = ffi.new("struct pollfd[1]", {{
        fd = socket.fd,
        events = table_to_flags(flags, POLL.lookup, bit.bor),
        revents = 0,
    }})
    local ok, err = socket.poll(pfd, 1, timeout or 0)
    if not ok then return ok, err end
    return flags_to_table(pfd[0].revents, POLL.lookup, bit.bor), ok
end

local function addrinfo_get_ip(self)
    if self.addrinfo.ai_addr == nil then
        return nil
    end
    local str = ffi.new("char[256]")
    local addr = assert(socket.inet_ntop(AF.lookup[self.family], ffi.cast("struct sockaddr_in*", self.addrinfo.ai_addr).sin_addr, str, ffi.sizeof(str)))
    return ffi.string(addr)
end

local function addrinfo_get_port(self)
    if self.addrinfo.ai_addr == nil then
        return nil
    end
    if self.family == "inet" then
        return socket.ntohs(ffi.cast("struct sockaddr_in*", self.addrinfo.ai_addr).sin_port)
    elseif self.family == "inet6" then
        return socket.ntohs(ffi.cast("struct sockaddr_in6*", self.addrinfo.ai_addr).sin6_port)
    end

    return nil, "unknown family " .. tostring(self.family)
end

local function addrinfo_to_table(res, host, service)
    local info = {}

    if res.ai_canonname ~= nil then
        info.canonical_name = ffi.string(res.ai_canonname)
    end

    info.host = host ~= "*" and host or nil
    info.service = service
    info.family = AF.reverse[res.ai_family]
    info.socket_type = SOCK.reverse[res.ai_socktype]
    info.protocol = IPPROTO.reverse[res.ai_protocol]
    info.flags = flags_to_table(res.ai_flags, AI.lookup, bit.band)
    info.addrinfo = res
    info.get_ip = addrinfo_get_ip
    info.get_port = addrinfo_get_port

    return info
end

function M.get_address_info(data)
    local hints

    if data.socket_type or data.protocol or data.flags or data.family then
        hints = ffi.new("struct addrinfo", {
            ai_family = data.family and AF.strict_lookup(data.family) or nil,
            ai_socktype = data.socket_type and SOCK.strict_lookup(data.socket_type) or nil,
            ai_protocol = data.protocol and IPPROTO.strict_lookup(data.protocol) or nil,
            ai_flags = data.flags and table_to_flags(data.flags, AI.lookup, bit.bor) or nil,
        })
    end

    local out = ffi.new("struct addrinfo*[1]")

    local ok, err = socket.getaddrinfo(
        data.host ~= "*" and data.host or nil,
        data.service and tostring(data.service) or nil,
        hints,
        out
    )

    if not ok then return ok, err end

    local tbl = {}

    local res = out[0]

    while res ~= nil do
        table.insert(tbl, addrinfo_to_table(res, data.host, data.service))

        res = res.ai_next
    end

    --ffi.C.freeaddrinfo(out[0])

    return tbl
end

function M.find_first_address(host, service, options)
    options = options or {}

    local info = {}
    info.host = host
    info.service = service

    info.family = options.family or "inet"
    info.socket_type = options.socket_type or "stream"
    info.protocol = options.protocol or "tcp"
    info.flags = options.flags

    if host == "*" then
        info.flags = info.flags or {}
        table.insert(info.flags, "passive")
    end

    local addrinfo, err = M.get_address_info(info)

    if not addrinfo then
        return nil, err
    end

    if not addrinfo[1] then
        return nil, "no addresses found (empty address info table)"
    end

    for _, v in ipairs(addrinfo) do
        if v.family == info.family and v.socket_type == info.socket_type and v.protocol == info.protocol then
            return v
        end
    end

    return addrinfo[1]
end


do
    local meta = {}
    meta.__index = meta

    function meta:__tostring()
        return string.format("socket[%s-%s-%s][%s]", self.family, self.socket_type, self.protocol, self.fd)
    end

    function M.create(family, socket_type, protocol)
        local fd, err, num = socket.create(AF.strict_lookup(family), SOCK.strict_lookup(socket_type), IPPROTO.strict_lookup(protocol))

        if not fd then return fd, err, num end

        return setmetatable({
            fd = fd,
            family = family,
            socket_type = socket_type,
            protocol = protocol,
            blocking = true,
        }, meta)
    end

    function meta:close()
        if self.on_close then
            self:on_close()
        end
        return socket.close(self.fd)
    end

    function meta:set_blocking(b)
        local ok, err, num = socket.blocking(self.fd, b)
        if ok then
            self.blocking = b
        end
        return ok, err, num
    end

    function meta:set_option(key, val, level)
        level = level or "socket"

        if type(val) == "boolean" then
            val = ffi.new("int[1]", val and 1 or 0)
        elseif type(val) == "number" then
            val = ffi.new("int[1]", val)
        elseif type(val) ~= "cdata" then
            error("unknown value type: " .. type(val))
        end

        local env = SO
        if level == "tcp" then
            env = TCP
        end

        return socket.setsockopt(self.fd, SOL.strict_lookup(level), env.strict_lookup(key), ffi.cast("void *", val), ffi.sizeof(val))
    end

    function meta:connect(host, service)
        local res

        if type(host) == "table" and host.addrinfo then
            res = host
        else
            local res_, err = M.find_first_address(host, service, {
                family = self.family,
                socket_type = self.socket_type,
                protocol = self.protocol
            })

            if not res_ then
                return res_, err
            end

            res = res_
        end

        local ok, err, num = socket.connect(self.fd, res.addrinfo.ai_addr, res.addrinfo.ai_addrlen)

        if not ok and not self.blocking then
            if timeout_messages[num] then
                self.timeout_connected = {host, service}
                return true
            end
        elseif self.on_connect then
            self:on_connect(host, service)
        end

        if not ok then
            return ok, err, num
        end

        return true
    end

    function meta:poll_connect()
        if self.on_connect and self.timeout_connected and self:is_connected() then
            local ok, err, num = self:on_connect(unpack(self.timeout_connected))
            self.timeout_connected = nil
            return ok, err, num
        end

        return nil, "timeout"
    end

    function meta:bind(host, service)
        if host == "*" then
            host = nil
        end

        if type(service) == "number" then
            service = tostring(service)
        end

        local res

        if type(host) == "table" and host.addrinfo then
            res = host
        else
            local res_, err = M.find_first_address(host, service, {
                family = self.family,
                socket_type = self.socket_type,
                protocol = self.protocol
            })

            if not res_ then
                return res_, err
            end

            res = res_
        end

        return socket.bind(self.fd, res.addrinfo.ai_addr, res.addrinfo.ai_addrlen)
    end

    function meta:listen(max_connections)
        max_connections = max_connections or e.SOMAXCONN
        return socket.listen(self.fd, max_connections)
    end

    function meta:accept()
        local address = ffi.new("struct sockaddr_in[1]")
        local fd, err = socket.accept(self.fd, ffi.cast("struct sockaddr *", address), ffi.new("unsigned int[1]", ffi.sizeof(address)))

        if fd ~= socket.INVALID_SOCKET then
            local client = setmetatable({
                fd = fd,
                family = "unknown",
                socket_type = "unknown",
                protocol = "unknown",
                blocking = true,
            }, meta)

            if self.debug then
                print(tostring(self), ": accept client: ", tostring(client))
            end

            return client
        end

        local err, num = socket.lasterror()

        if not self.blocking and timeout_messages[num] then
            return nil, "timeout", num
        end

        if self.debug then
            print(tostring(self), ": accept error", num, ":", err)
        end

        return nil, err, num
    end

    function meta:is_connected()
        local ip, service, num = self:get_peer_name()
        local ip2, service2, num2 = self:get_name()

        if not ip and (num == errno.ECONNRESET or num == errno.ENOTSOCK) then
            return false, service, num
        end

        if ffi.os == "Windows" then
            return ip ~= "0.0.0.0" and ip2 ~= "0.0.0.0" and service ~= 0 and service2 ~= 0
        else
            return ip and ip2 and service ~= 0 and service2 ~= 0
        end
    end

    function meta:get_peer_name()
        local data = ffi.new("struct sockaddr_in")
        local len = ffi.new("unsigned int[1]", ffi.sizeof(data))

        local ok, err, num = socket.getpeername(self.fd, ffi.cast("struct sockaddr *", data), len)
        if not ok then return ok, err, num end

        return ffi.string(socket.inet_ntoa(data.sin_addr)), socket.ntohs(data.sin_port)
    end

    function meta:get_name()
        local data = ffi.new("struct sockaddr_in")
        local len = ffi.new("unsigned int[1]", ffi.sizeof(data))

        local ok, err, num = socket.getsockname(self.fd, ffi.cast("struct sockaddr *", data), len)
        if not ok then return ok, err, num end

        return ffi.string(socket.inet_ntoa(data.sin_addr)), socket.ntohs(data.sin_port)
    end

    local default_flags = 0

    if ffi.os ~= "Windows" then
        default_flags = e.MSG_NOSIGNAL
    end

    function meta:send_to(addr, data, flags)
        return self:send(data, flags, addr)
    end

    function meta:send(data, flags, addr)
        flags = flags or default_flags

        if self.on_send then
            return self:on_send(data, flags)
        end

        local len, err, num

        if addr then
            len, err, num = socket.sendto(self.fd, data, #data, flags, addr.addrinfo.ai_addr, addr.addrinfo.ai_addrlen)
        else
            len, err, num = socket.send(self.fd, data, #data, flags)
        end

        if not len then
            return len, err, num
        end

        if len > 0 then
            return len
        end
    end

    function meta:receive_from(address, size, flags)
        local src_addr
        local src_addr_size

        if not address then
            src_addr = ffi.new("struct sockaddr_in[1]")
            src_addr_size = ffi.sizeof("struct sockaddr_in")
        else
            src_addr = address.addrinfo.ai_addr
            src_addr_size = address.addrinfo.ai_addrlen
        end

        return self:receive(size, flags, src_addr, src_addr_size)
    end

    function meta:receive(size, flags, src_address, address_len)
        size = size or 64000
        local buff = ffi.new("char[?]", size)

        if self.on_receive then
            return self:on_receive(buff, size, flags)
        end

        local len, err, num
        local len_res

        if src_address then
            len_res = ffi.new("int[1]", address_len)
            len, err, num = socket.recvfrom(self.fd, buff, ffi.sizeof(buff), flags or 0, ffi.cast("struct sockaddr *", src_address), len_res)
        else
            len, err, num = socket.recv(self.fd, buff, ffi.sizeof(buff), flags or 0)
        end

        if num == errno.ECONNRESET then
            self:close()
            if self.debug then
                print(tostring(self), ": closed")
            end

            return nil, "closed", num
        end

        if not len then
            if not self.blocking and timeout_messages[num] then
                return nil, "timeout", num
            end

            if self.debug then
                print(tostring(self), " error", num, ":", err)
            end

            return len, err, num
        end

        if len > 0 then
            if self.debug then
                print(tostring(self), ": received ", len, " bytes")
            end

            if src_address then
                return ffi.string(buff, len), {
                    addrinfo = {
                        ai_addr = ffi.cast("struct sockaddr *", src_address),
                        ai_addrlen = len_res[0],
                    },
                    family = self.family,
                    get_port = addrinfo_get_port,
                    get_ip = addrinfo_get_ip,
                }
            end

            return ffi.string(buff, len)
        end

        return nil, err, num
    end
end

function M.bind(host, service)
    local info, err = M.find_first_address(host, service, {
        family = "inet",
        socket_type = "stream",
        protocol = "tcp",
        flags = {"passive"},
    })

    if not info then
        return info, err
    end

    local server, err, num = M.create(info.family, info.socket_type, info.protocol)

    if not server then
        return server, err, num
    end

    server:set_option("reuseaddr", 1)

    local ok, err, num = server:bind(info)

    if not ok then
        return ok, err, num
    end

    server:set_option("sndbuf", 65536)
    server:set_option("rcvbuf", 65536)

    return server
end

return M
end
end

do
local _ENV = _ENV
package.preload[ "modules.log" ] = function( ... ) local arg = _G.arg;
local log = {};

local json;
local data;
local text_source_handler;

local obslua = obslua;
local require = require;
local script_path = script_path;
local print = print;

local file_output_name = "obs-stats_" .. os.date("%d.%m.%Y") .. ".log";

local log_file_formatting = [[
[$date_time] Streaming duration: $streaming_duration, Recording Duration: $recording_duration, Streaming status: $streaming_status, Recording_status: $recording_status, Encoder: $encoder, Output Mode: $output_mode, Canvas Resolution: $canvas_resolution, Output Resolution: $output_resolution, Missed Frames: $missed_frames/$missed_total_frames ($missed_percents%), Skipped Frames: $skipped_frames/$skipped_total_frames ($skipped_percents%), Dropped Frames: $dropped_frames/$dropped_total_frames ($dropped_percents%), Congestion: $congestion% (avg. $average_congestion%), Memory Usage: $memory_usage MB, CPU Cores: $cpu_cores, CPU Usage: $cpu_usage%, Frame Time: $average_frame_time ms, FPS: $fps/$target_fps (avg. $average_fps), Bitrate: $bitrate kb/s, Audio Bitrate: $audio_bitrate kb/s, Recording bitrate: $recording_bitrate kb/s
]];

function log.to_file()
	--print("Log to " .. file_output_name);

	local data_log = text_source_handler.format_text(log_file_formatting);
	--local data_log = json.encode(data.stats, { indent = false }) .. "\n";

	local script_path_ = script_path();
	local output_path = script_path_ .. file_output_name;

	local log_file = io.open(output_path, "a");
	log_file:write(data_log);
	log_file:close();
end

function log.init_module()
	json = require("modules.json");
	data = require("modules.data");
	text_source_handler = require("modules.text_source_handler");
end

return log;
end
end

do
local _ENV = _ENV
package.preload[ "modules.output" ] = function( ... ) local arg = _G.arg;
local output = {};

local json;
local data;

local obslua = obslua;
local require = require;
local script_path = script_path;
local print = print;

local file_output_name = "obs-stats.json";

function output.to_json()
	--print("Output to " .. file_output_name);

	local data_json = json.encode(data.stats, { indent = true });

	local script_path_ = script_path();
	local output_path = script_path_ .. file_output_name;

	obslua.os_quick_write_utf8_file(output_path, data_json, #data_json, false);
end

function output.init_module()
	json = require("modules.json");
	data = require("modules.data");
end

return output;
end
end

do
local _ENV = _ENV
package.preload[ "modules.profile_handler" ] = function( ... ) local arg = _G.arg;
local profile_handler = {};

local data;

local obslua = obslua;
local print = print;
local tonumber = tonumber;
local require = require;

function profile_handler.read_config()
	local profile = obslua.obs_frontend_get_current_profile():gsub("[^%w_ ]", ""):gsub("%s", "_");
	
	local profile_relative_path = "obs-studio\\basic\\profiles\\" .. profile .. "\\basic.ini";
	
	-- char profile_path[512];
	local profile_path = "                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                ";
	obslua.os_get_abs_path("..\\..\\config\\" .. profile_relative_path, profile_path, #profile_path);
	
	if not obslua.os_file_exists(profile_path) then	
		obslua.os_get_config_path(profile_path, #profile_path, profile_relative_path);
	
		if not obslua.os_file_exists(profile_path) then	
			print("Profile Config File not found.");
			return;
		end
	end

	local config_text = obslua.os_quick_read_utf8_file(profile_path);

	if config_text == nil then 
		print("Couldn't read Profile Config File.");
		return;
	end
	
	print("Profile Config loaded: " ..  profile_path);
	
	local config = profile_handler.parse_ini(config_text);
	
	if config.Output == nil then
		data.update_output_mode(nil);
	else
		data.update_output_mode(config.Output.Mode);
	end

	if config.Video == nil then
		data.update_target_fps(nil);
	else
		data.update_canvas_resolution(config.Video.BaseCX, config.Video.BaseCY);
		data.update_output_resolution(config.Video.OutputCX, config.Video.OutputCY);
		data.update_target_fps(config.Video.FPSCommon);
	end

	if data.stats.output_mode == data.output_modes.simple then
		if config.SimpleOutput == nil then
			data.update_audio_bitrate(nil);
		else
			data.update_encoder(config.SimpleOutput.StreamEncoder);
			data.update_audio_bitrate(config.SimpleOutput.ABitrate);
		end
	else
		if config.AdvOut == nil then
			data.update_audio_bitrate(nil);
		else
			data.update_encoder(config.AdvOut.Encoder);
			data.update_audio_bitrate(config.AdvOut.Track1Bitrate);
		end
	end
end

function profile_handler.parse_ini(ini_text)
	local data = {};
	local section;

	for line in ini_text:gmatch("[^\r\n]+") do
		local tempSection = line:match('^%[([^%[%]]+)%]$');

		if tempSection then
			section = tonumber(tempSection) and tonumber(tempSection) or tempSection;
			data[section] = data[section] or {};
		end

		local param, value = line:match('^([%w|_]+)%s-=%s-(.+)$');

		if param and value ~= nil then

			if tonumber(value) then
				value = tonumber(value);
			elseif value == 'true' then
				value = true;
			elseif value == 'false' then
				value = false;
			end

			if tonumber(param) then
				param = tonumber(param);
			end

			data[section][param] = value;
		end
	end
	return data;
end

function profile_handler.init_module()
	data = require("modules.data");
end

return profile_handler;
end
end

do
local _ENV = _ENV
package.preload[ "modules.script_handler" ] = function( ... ) local arg = _G.arg;
local script_handler = {};

local data;
local data_format;
local bot;
local output;
local log;
local profile_handler;
local text_source_handler;

local obslua = obslua;
local ipairs = ipairs;
local print = print;
local require = require;

local description = [[
<center><h2>OBS Stats on Stream v2.1</h2></center>
<center>Made by <a href="https://twitch.tv/greencomfytea">GreenComfyTea</a> | 2024</center>
<center><a href="https://linktr.ee/greencomfytea">Socials</a> | <a href="https://ko-fi.com/greencomfytea">Buy me a tea</a> | <a href="https://streamelements.com/greencomfytea/tip">Donate</a></center>
<center><p>Shows OBS Stats (like Bitrate, Dropped Frames and more) on Stream and/or in Twitch Chat.</p></center>
<center><a href="https://twitchapps.com/tmi/">Twitch Chat OAuth Password Generator</a></center>
<center><a href="https://github.com/GreenComfyTea/OBS-Stats-on-Stream/blob/main/Text-Formatting-Variables.md">Text Formatting Variables</a> | <a href="https://github.com/GreenComfyTea/OBS-Stats-on-Stream/blob/main/Bot-Commands.md">Bot commands</a></center>
<hr/>
]];

local my_settings = nil;

script_handler.is_script_enabled = true;
script_handler.timer_delay = 1000;
script_handler.text_source = "";
script_handler.text_formatting = "";

script_handler.is_output_to_file_enabled = false;
script_handler.is_logging_enabled = false;

script_handler.is_bot_enabled = true;
script_handler.bot_delay = 2000;

script_handler.bot_password = "";
script_handler.bot_nickname = "justinfan4269";

script_handler.channel_nickname = "";

script_handler.is_timer_on = false;

function script_handler.tick()
	--print("Tick");

	if not script_handler.is_script_enabled then
		return;
	end

	data.update();
	data_format.update();
	text_source_handler.update();

	if script_handler.is_output_to_file_enabled then
		output.to_json();
	end

	if script_handler.is_logging_enabled then
		log.to_file();
	end
end

function script_handler.reset_formatting(properties, property)
	script_handler.text_formatting = text_source_handler.default_formatting;

	obslua.obs_data_set_string(my_settings, "text_formatting", text_source_handler.default_formatting);
	obslua.obs_properties_apply_settings(properties, my_settings);

	return true;
end

function script_handler.on_event(event)
	if event == obslua.OBS_FRONTEND_EVENT_STREAMING_STARTED then
		data.update_streaming_status(true);
	elseif event == obslua.OBS_FRONTEND_EVENT_STREAMING_STOPPED then
		data.update_streaming_status(false);
	elseif event == obslua.OBS_FRONTEND_EVENT_RECORDING_STARTED then
		data.update_recording_status(true, false);
	elseif event == obslua.OBS_FRONTEND_EVENT_RECORDING_STOPPED then
		data.update_recording_status(false, false);
	elseif event == obslua.OBS_FRONTEND_EVENT_RECORDING_PAUSED then
		data.update_recording_status(true, true);
	elseif event == obslua.OBS_FRONTEND_EVENT_RECORDING_UNPAUSED then
		data.update_recording_status(true, false);
	elseif event == obslua.OBS_FRONTEND_EVENT_FINISHED_LOADING
	or event == obslua.OBS_FRONTEND_EVENT_PROFILE_CHANGED
	or event == obslua.OBS_FRONTEND_EVENT_PROFILE_LIST_CHANGED
	or event == obslua.OBS_FRONTEND_EVENT_STREAMING_STARTING
	or event == obslua.OBS_FRONTEND_EVENT_STREAMING_STOPPING
	or event == obslua.OBS_FRONTEND_EVENT_RECORDING_STARTING
	or event == obslua.OBS_FRONTEND_EVENT_RECORDING_STOPPING then
		profile_handler.read_config();
	end
end

function script_properties()
	local properties = obslua.obs_properties_create();

	local enable_script_property = obslua.obs_properties_add_bool(properties, "is_script_enabled", "Enable Script");
	local enable_bot_property = obslua.obs_properties_add_bool(properties, "is_bot_enabled", "Enable Bot");

	local timer_delay_property = obslua.obs_properties_add_int(properties, "timer_delay", "Update Delay (ms)", 100, 10000, 100);
	obslua.obs_property_set_long_description(timer_delay_property, "Determines how often the data will update.");
	
	local bot_delay_property = obslua.obs_properties_add_int(properties, "bot_delay", "Bot Delay (ms)", 500, 5000, 100);
	obslua.obs_property_set_long_description(bot_delay_property, "Determines how often the bot will read chat and write to it.");

	local bot_nickname_property = obslua.obs_properties_add_text(properties, "bot_nickname", "Bot Nickname", obslua.OBS_TEXT_DEFAULT);
	obslua.obs_property_set_long_description(bot_nickname_property, "Nickname of your bot.");
	
	local bot_oauth_property = obslua.obs_properties_add_text(properties, "bot_password", "Bot OAuth Password", obslua.OBS_TEXT_PASSWORD);
	obslua.obs_property_set_long_description(bot_oauth_property, "Format: oauth:xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx. Visit https://twitchapps.com/tmi/ to get the AOuth Password for the bot (you must login to twitch.tv accordingly.");

	local channel_nickname_property = obslua.obs_properties_add_text(properties, "channel_nickname", "Channel Nickname", obslua.OBS_TEXT_DEFAULT);
	obslua.obs_property_set_long_description(channel_nickname_property, "Nickname of your channel for bot to join. If empty bot will join his own chat.");

	obslua.obs_properties_add_button(properties, "recconect_button", "Reconnect...", bot.reconnect);

	local text_source_property = obslua.obs_properties_add_list(properties, "text_source", "Text Source", obslua.OBS_COMBO_TYPE_EDITABLE, obslua.OBS_COMBO_FORMAT_STRING);

	local sources = obslua.obs_enum_sources();
	if sources ~= nil then
		for _, source in ipairs(sources) do
			local source_id = obslua.obs_source_get_id(source);
			if source_id == "text_gdiplus_v2" or source_id == "text_ft2_source_v2" then
				local name = obslua.obs_source_get_name(source);
				obslua.obs_property_list_add_string(text_source_property, name, name);
			end
		end
	end
	obslua.source_list_release(sources);
	
	obslua.obs_property_set_long_description(text_source_property, "Text source that will be used to display the data.");

	obslua.obs_properties_add_text(properties, "text_formatting", "Text Formatting", obslua.OBS_TEXT_MULTILINE);
	obslua.obs_properties_add_button(properties, "reset_formatting_button", "Reset Formatting", script_handler.reset_formatting);

	obslua.obs_properties_apply_settings(properties, my_settings);

	local enable_output_to_file_property = obslua.obs_properties_add_bool(properties, "is_output_to_file_enabled", "Output to File");
	local enable_logging_property = obslua.obs_properties_add_bool(properties, "is_logging_enabled", "Enable Logging");

	return properties;
end

function script_defaults(settings)
	my_settings = settings;

	obslua.obs_data_set_default_bool(settings, "is_script_enabled", true);
	obslua.obs_data_set_default_bool(settings, "is_bot_enabled", true);

	obslua.obs_data_set_default_int(settings, "timer_delay", 1000);
	obslua.obs_data_set_default_int(settings, "bot_delay", 2000);

	obslua.obs_data_set_default_string(settings, "bot_nickname", "");
	obslua.obs_data_set_default_string(settings, "bot_password", "");
	obslua.obs_data_set_default_string(settings, "channel_nickname", "");

	obslua.obs_data_set_default_string(settings, "text_source", "");
	obslua.obs_data_set_default_string(settings, "text_formatting", text_source_handler.default_formatting);

	obslua.obs_data_set_default_bool(settings, "is_output_to_file_enabled", false);
	obslua.obs_data_set_default_bool(settings, "is_logging_enabled", false);

	print("Settings were reset.");
end

function script_update(settings)
	my_settings = settings;

	local previous_is_script_enabled = script_handler.is_script_enabled;
	local previous_is_bot_enabled = script_handler.is_bot_enabled;
	local previous_bot_delay = script_handler.bot_delay;
	local previous_bot_nickname = script_handler.bot_nickname;
	local previous_bot_password = script_handler.bot_password;
	local previous_channel_nickname = script_handler.channel_nickname;

	script_handler.is_script_enabled = obslua.obs_data_get_bool(settings, "is_script_enabled");
	script_handler.is_bot_enabled = obslua.obs_data_get_bool(settings, "is_bot_enabled");

	script_handler.timer_delay = obslua.obs_data_get_int(settings, "timer_delay");
	script_handler.bot_delay = obslua.obs_data_get_int(settings, "bot_delay");

	script_handler.bot_nickname = obslua.obs_data_get_string(settings, "bot_nickname");
	script_handler.bot_password = obslua.obs_data_get_string(settings, "bot_password");
	script_handler.channel_nickname = obslua.obs_data_get_string(settings, "channel_nickname");

	script_handler.text_source = obslua.obs_data_get_string(settings, "text_source");
	script_handler.text_formatting = obslua.obs_data_get_string(settings, "text_formatting");

	script_handler.is_output_to_file_enabled = obslua.obs_data_get_bool(settings, "is_output_to_file_enabled");
	script_handler.is_logging_enabled = obslua.obs_data_get_bool(settings, "is_logging_enabled");

	local reconnect_bot = script_handler.is_script_enabled ~= previous_is_script_enabled
		or script_handler.is_bot_enabled ~= previous_is_bot_enabled
		or script_handler.bot_delay ~= previous_bot_delay
		or script_handler.bot_nickname ~= previous_bot_nickname
		or script_handler.bot_password ~= previous_bot_password
		or script_handler.channel_nickname ~= previous_channel_nickname;

	data.start_cpu_usage_info();

	if script_handler.is_timer_on then
		script_handler.is_timer_on = false;
		obslua.timer_remove(script_handler.tick);
		
		print("Timer removed.");
	end

	if reconnect_bot then
		bot.close_socket();
	end
	
	print("Settings were updated.");

	profile_handler.read_config();
	data.update_cores_on_script_settings_changed();
	data.update_streaming_status_on_script_settings_changed();
	data.update_recording_status_on_script_settings_changed();

	if script_handler.is_script_enabled then
		script_handler.is_timer_on = true;
		obslua.timer_add(script_handler.tick, script_handler.timer_delay);
		print("Timer added.");

		if reconnect_bot then
			bot.init_socket();
		end
	end
end

function script_description()
	return description;
end

function script_load(settings)
	obslua.obs_frontend_add_event_callback(script_handler.on_event);
	print("Script is loaded.");
end

function script_unload()
	-- not working???

	--if is_script_enabled then
	--	is_timer_on = false;
	--	obslua.timer_remove(data.tick);
	--	print("Timer removed.");
	--end

	bot.close_socket_on_unload();

	print("[OBS-Stats-on-Stream.lua] Script unloaded.");
end

function script_handler.init_module()
	data = require("modules.data");
	data_format = require("modules.data_format");
	output = require("modules.output");
	log = require("modules.log");
	profile_handler = require("modules.profile_handler");
	text_source_handler = require("modules.text_source_handler");
	bot = require("modules.bot");
end

return script_handler;
end
end

do
local _ENV = _ENV
package.preload[ "modules.text_source_handler" ] = function( ... ) local arg = _G.arg;
local text_source_handler = {};

local data_format;
local script_handler;
local utils;

local obslua = obslua;
local print = print;
local os = os;
local tostring = tostring;
local string = string;
local require = require;

text_source_handler.default_formatting = [[
Missed Frames: $missed_frames/$missed_total_frames ($missed_percents%)
Skipped Frames: $skipped_frames/$skipped_total_frames ($skipped_percents%)
Dropped Frames: $dropped_frames/$dropped_total_frames ($dropped_percents%)
Congestion: $congestion% (avg. $average_congestion%)
Memory Usage: $memory_usage MB
CPU Usage: $cpu_usage%
Frame Time: $average_frame_time ms
FPS: $fps/$target_fps (avg. $average_fps)
Bitrate: $bitrate kb/s
]];

function text_source_handler.update()
	--print("Text Source Update Tick.");

	local text_source_name = script_handler.text_source;

	if text_source_name == nil or text_source_name == "" then
		print("Text Source not specified.");
		return;
	end

	local source = obslua.obs_get_source_by_name(text_source_name);
	if source == nil then
		print("Text Source not found.");
	end

	-- Make a string for display in a text source
	local formatted_text = text_source_handler.format_text(script_handler.text_formatting);

	local settings = obslua.obs_data_create();
	obslua.obs_data_set_string(settings, "text", formatted_text);
	obslua.obs_source_update(source, settings);
	obslua.obs_source_release(source);
	obslua.obs_data_release(settings);
end

function text_source_handler.format_text(text_formatting)
	local formatted_text = text_formatting;
	local formatted_stats = data_format.stats;

	formatted_text = formatted_text:gsub("$date_time", formatted_stats.date_time);
	formatted_text = formatted_text:gsub("$date", formatted_stats.date);
	formatted_text = formatted_text:gsub("$time", formatted_stats.time);

	formatted_text = formatted_text:gsub("$encoder", formatted_stats.encoder);
	formatted_text = formatted_text:gsub("$output_mode", formatted_stats.output_mode);

	formatted_text = formatted_text:gsub("$canvas_width", formatted_stats.canvas_width);
	formatted_text = formatted_text:gsub("$canvas_height", formatted_stats.canvas_height);
	formatted_text = formatted_text:gsub("$canvas_resolution", formatted_stats.canvas_resolution);

	formatted_text = formatted_text:gsub("$output_width", formatted_stats.output_width);
	formatted_text = formatted_text:gsub("$output_height", formatted_stats.output_height);
	formatted_text = formatted_text:gsub("$output_resolution", formatted_stats.output_resolution);

	formatted_text = formatted_text:gsub("$missed_frames", formatted_stats.missed_frames);
	formatted_text = formatted_text:gsub("$missed_total_frames", formatted_stats.total_missed_frames);
	formatted_text = formatted_text:gsub("$total_missed_frames", formatted_stats.total_missed_frames);
	formatted_text = formatted_text:gsub("$missed_percents", formatted_stats.missed_percents);

	formatted_text = formatted_text:gsub("$skipped_frames", formatted_stats.skipped_frames);
	formatted_text = formatted_text:gsub("$skipped_total_frames", formatted_stats.total_skipped_frames);
	formatted_text = formatted_text:gsub("$total_skipped_frames", formatted_stats.total_skipped_frames);
	formatted_text = formatted_text:gsub("$skipped_percents", formatted_stats.skipped_percents);

	formatted_text = formatted_text:gsub("$dropped_frames", formatted_stats.dropped_frames);
	formatted_text = formatted_text:gsub("$dropped_total_frames", formatted_stats.total_dropped_frames);
	formatted_text = formatted_text:gsub("$total_dropped_frames", formatted_stats.total_dropped_frames);
	formatted_text = formatted_text:gsub("$dropped_percents", formatted_stats.dropped_percents);

	formatted_text = formatted_text:gsub("$congestion", formatted_stats.congestion);
	formatted_text = formatted_text:gsub("$average_congestion", formatted_stats.average_congestion);

	formatted_text = formatted_text:gsub("$average_frame_time", formatted_stats.average_frame_time);
	formatted_text = formatted_text:gsub("$fps", formatted_stats.fps);
	formatted_text = formatted_text:gsub("$target_fps", formatted_stats.target_fps);
	formatted_text = formatted_text:gsub("$average_fps", formatted_stats.average_fps);
	
	formatted_text = formatted_text:gsub("$memory_usage", formatted_stats.memory_usage);
	formatted_text = formatted_text:gsub("$cpu_physical_cores", formatted_stats.cpu_physical_cores);
	formatted_text = formatted_text:gsub("$cpu_logical_cores", formatted_stats.cpu_logical_cores);
	formatted_text = formatted_text:gsub("$cpu_cores", formatted_stats.cpu_cores);
	formatted_text = formatted_text:gsub("$cpu_usage", formatted_stats.cpu_usage);

	formatted_text = formatted_text:gsub("$audio_bitrate", formatted_stats.audio_bitrate);
	formatted_text = formatted_text:gsub("$recording_bitrate", formatted_stats.recording_bitrate);
	formatted_text = formatted_text:gsub("$bitrate", formatted_stats.bitrate);

	formatted_text = formatted_text:gsub("$streaming_total_seconds", formatted_stats.streaming_total_seconds);
	formatted_text = formatted_text:gsub("$streaming_total_minutes", formatted_stats.streaming_total_minutes);
	formatted_text = formatted_text:gsub("$streaming_hours", formatted_stats.streaming_hours);
	formatted_text = formatted_text:gsub("$streaming_minutes", formatted_stats.streaming_minutes);
	formatted_text = formatted_text:gsub("$streaming_seconds", formatted_stats.streaming_seconds);
	formatted_text = formatted_text:gsub("$streaming_duration", formatted_stats.streaming_duration);

	formatted_text = formatted_text:gsub("$recording_total_seconds", formatted_stats.recording_total_seconds);
	formatted_text = formatted_text:gsub("$recording_total_minutes", formatted_stats.recording_total_minutes);
	formatted_text = formatted_text:gsub("$recording_hours", formatted_stats.recording_hours);
	formatted_text = formatted_text:gsub("$recording_minutes", formatted_stats.recording_minutes);
	formatted_text = formatted_text:gsub("$recording_seconds", formatted_stats.recording_seconds);
	formatted_text = formatted_text:gsub("$recording_duration", formatted_stats.recording_duration);

	formatted_text = formatted_text:gsub("$streaming_status", formatted_stats.streaming_status);
	formatted_text = formatted_text:gsub("$recording_status", formatted_stats.recording_status);

	return formatted_text;
end

function text_source_handler.init_module()
	data_format = require("modules.data_format");
	script_handler = require("modules.script_handler");
	utils = require("modules.utils");
end

return text_source_handler;
end
end

do
local _ENV = _ENV
package.preload[ "modules.utils" ] = function( ... ) local arg = _G.arg;
local utils = {};

local next = next;
local type = type;
local pairs = pairs;
local string = string;
local table = table;
local tostring = tostring;
local require = require;
local math = math;

function utils.table_tostring(table_)
	if type(table_) == "number" or type(table_) == "boolean" or type(table_) == "string" then
		return tostring(table_);
	end

	if utils.is_table_empty(table_) then
		return "{}"; 
	end

	local cache = {};
	local stack = {};
	local output = {};
    local depth = 1;
    local output_str = "{\n";

    while true do
        local size = 0;
        for k,v in pairs(table_) do
            size = size + 1;
        end

        local cur_index = 1;
        for k,v in pairs(table_) do
            if cache[table_] == nil or cur_index >= cache[table_] then

                if string.find(output_str, "}", output_str:len()) then
                    output_str = output_str .. ",\n";
                elseif not string.find(output_str, "\n", output_str:len()) then
                    output_str = output_str .. "\n";
                end

                -- This is necessary for working with HUGE tables otherwise we run out of memory using concat on huge strings
                table.insert(output,output_str);
                output_str = "";

                local key;
                if type(k) == "number" or type(k) == "boolean" then
                    key = "[" .. tostring(k) .. "]";
                else
                    key = "['" .. tostring(k) .. "']";
                end

                if type(v) == "number" or type(v) == "boolean" then
                    output_str = output_str .. string.rep('\t', depth) .. key .. " = "..tostring(v);
                elseif type(v) == "table" then
                    output_str = output_str .. string.rep('\t', depth) .. key .. " = {\n";
                    table.insert(stack, table_);
                    table.insert(stack, v);
                    cache[table_] = cur_index + 1;
                    break;
                else
                    output_str = output_str .. string.rep('\t', depth) .. key .. " = '" .. tostring(v) .. "'";
                end

                if cur_index == size then
                    output_str = output_str .. "\n" .. string.rep('\t', depth - 1) .. "}";
                else
                    output_str = output_str .. ",";
                end
            else
                -- close the table
                if cur_index == size then
                    output_str = output_str .. "\n" .. string.rep('\t', depth - 1) .. "}";
                end
            end

            cur_index = cur_index + 1;
        end

        if size == 0 then
            output_str = output_str .. "\n" .. string.rep('\t', depth - 1) .. "}";
        end

        if #stack > 0 then
            table_ = stack[#stack];
            stack[#stack] = nil;
            depth = cache[table_] == nil and depth + 1 or depth - 1;
        else
            break
        end
    end

    -- This is necessary for working with HUGE tables otherwise we run out of memory using concat on huge strings
    table.insert(output, output_str);
    output_str = table.concat(output);

    return output_str;
end

function utils.table_tostringln(table_)
	return "\n" .. utils.table_tostring(table_);
end

function utils.is_table_empty(table_)
	return next(table_) == nil;
end

function utils.is_NaN(value)
	return tostring(value) == tostring(0/0);
end

function utils.round(value)
	return math.floor(value + 0.5);
end

function utils.trim(str)
	return str:match("^%s*(.-)%s*$");
end

function utils.starts_with(str, pattern)
	return str:find("^".. pattern) ~= nil;
end

function utils.init_module()
end

return utils;
end
end

local require = require;
local print = print;

local json = require("modules.json");
local ljsocket = require("modules.ljsocket");
local bot = require("modules.bot");
local data_format = require("modules.data_format");
local data = require("modules.data");
local output = require("modules.output");
local log = require("modules.log");
local profile_handler = require("modules.profile_handler");
local script_handler = require("modules.script_handler");
local text_source_handler = require("modules.text_source_handler");
local utils = require("modules.utils");

bot.init_module();
data_format.init_module();
data.init_module();
output.init_module();
log.init_module();
profile_handler.init_module();
script_handler.init_module();
text_source_handler.init_module();
utils.init_module();